"""Search helpers for SmarterMail logs."""

from __future__ import annotations

from array import array
from collections import OrderedDict
from dataclasses import dataclass, field
from difflib import SequenceMatcher
from pathlib import Path
import re
from threading import Lock
import time
from typing import Any, Callable, List, Protocol, Tuple

from .log_kinds import (
    KIND_ADMINISTRATIVE,
    KIND_DELIVERY,
    KIND_IMAP,
    KIND_IMAP_RETRIEVAL,
    KIND_POP,
    KIND_SMTP,
    is_search_ungrouped_kind,
    normalize_kind,
)
from .log_parsers import (
    starts_with_timestamp,
)
from .search_modes import (
    DEFAULT_FUZZY_THRESHOLD,
    MODE_FUZZY,
    MODE_LITERAL,
    MODE_REGEX,
    MODE_WILDCARD,
    normalize_fuzzy_threshold,
    normalize_search_mode,
    wildcard_to_regex,
)

def _load_rapidfuzz_fuzz() -> Any:
    try:
        from rapidfuzz import fuzz
    except Exception:  # pragma: no cover - optional dependency
        return None
    return fuzz


_rapidfuzz_fuzz: Any = _load_rapidfuzz_fuzz()


_FUZZY_ANCHOR_LIMIT = 120
_FUZZY_STRIDE_DIVISOR = 6
_TIME_PATTERN = r"\d{2}:\d{2}:\d{2}(?:\.\d{3})?"
_SMTP_OWNER_PATTERN = re.compile(
    rf"^(?:{_TIME_PATTERN}) \[[^\]]+\]\[([^\]]+)\] "
)
_DELIVERY_OWNER_PATTERN = re.compile(
    rf"^(?:{_TIME_PATTERN}) \[([^\]]+)\] "
)
_IMAP_RETRIEVAL_OWNER_PATTERN = re.compile(
    rf"^(?:{_TIME_PATTERN}) \[([^\]]*)\] \[[^\]]*\] "
)
_ADMIN_OWNER_PATTERN = re.compile(
    rf"^(?P<time>{_TIME_PATTERN}) \[(?P<ip>[^\]]*)\] "
)
_ADMIN_TRAILING_OWNER_PATTERN = re.compile(
    rf"^\[(?P<ip>[^\]]+)\]\s+.*\s+(?P<time>{_TIME_PATTERN})$"
)
MATERIALIZATION_AUTO = "auto"
MATERIALIZATION_SINGLE_PASS = "single-pass"
MATERIALIZATION_TWO_PASS = "two-pass"
SUPPORTED_MATERIALIZATION_MODES = (
    MATERIALIZATION_AUTO,
    MATERIALIZATION_SINGLE_PASS,
    MATERIALIZATION_TWO_PASS,
)
_AUTO_SAMPLE_LINES = 10000
_AUTO_MAX_LINE_MATCH_RATIO = 0.002
_AUTO_MAX_OWNER_MATCH_RATIO = 0.04
_INDEX_CACHE_MAX_BYTES = 256 * 1024 * 1024
_PROGRESS_REPORT_BYTES = 4 * 1024 * 1024
_PROGRESS_REPORT_SECONDS = 0.2


@dataclass
class Conversation:
    """Log conversation grouped by message identifier."""

    message_id: str
    lines: List[str]
    first_line_number: int


@dataclass
class SmtpSearchResult:
    """Aggregate information about a search."""

    term: str
    log_path: Path
    conversations: List[Conversation]
    total_lines: int
    orphan_matches: List[Tuple[int, str]]
    matching_rows: List[Tuple[int, str]] = field(default_factory=list)

    @property
    def total_conversations(self) -> int:
        """Return the number of grouped conversations in ``conversations``."""

        return len(self.conversations)


class SearchFunction(Protocol):
    """Callable signature used by all search handlers."""

    def __call__(
        self,
        log_path: Path,
        term: str,
        *,
        mode: str = MODE_LITERAL,
        fuzzy_threshold: float = DEFAULT_FUZZY_THRESHOLD,
        ignore_case: bool = True,
        materialization: str = MATERIALIZATION_AUTO,
        use_index_cache: bool = False,
        progress_callback: "SearchProgressCallback | None" = None,
        match_callback: "SearchMatchCallback | None" = None,
    ) -> SmtpSearchResult: ...


SearchProgressCallback = Callable[[int, int], None]
SearchMatchCallback = Callable[[int, str], None]


@dataclass
class _SearchProgress:
    callback: SearchProgressCallback
    total_bytes: int
    scanned_bytes: int = 0
    next_report_bytes: int = _PROGRESS_REPORT_BYTES
    last_report_at: float = 0.0


@dataclass(frozen=True)
class _OwnerLineIndex:
    key: str
    signature: tuple[int, int]
    owner_codes: array
    owner_ids: list[str]
    owner_first_lines: list[int]
    line_count: int
    size_bytes: int


_INDEX_CACHE: "OrderedDict[tuple[str, str, int, int], _OwnerLineIndex]" = (
    OrderedDict()
)
_INDEX_CACHE_BYTES = 0
_INDEX_CACHE_LOCK = Lock()


def _safe_file_size(log_path: Path) -> int:
    try:
        return log_path.stat().st_size
    except OSError:
        return 0


def _start_search_progress(
    log_path: Path,
    callback: SearchProgressCallback | None,
    *,
    expected_passes: int,
) -> _SearchProgress | None:
    if callback is None:
        return None
    file_size = _safe_file_size(log_path)
    total_bytes = max(1, file_size * max(1, expected_passes))
    progress = _SearchProgress(
        callback=callback,
        total_bytes=total_bytes,
    )
    callback(0, total_bytes)
    return progress


def _advance_search_progress(
    progress: _SearchProgress | None,
    raw_line: str,
) -> None:
    if progress is None:
        return
    progress.scanned_bytes += len(raw_line)
    if progress.scanned_bytes < progress.next_report_bytes:
        return
    now = time.perf_counter()
    if now - progress.last_report_at < _PROGRESS_REPORT_SECONDS:
        return
    scanned = min(progress.scanned_bytes, progress.total_bytes)
    progress.callback(scanned, progress.total_bytes)
    progress.next_report_bytes = (
        progress.scanned_bytes + _PROGRESS_REPORT_BYTES
    )
    progress.last_report_at = now


def _finish_search_progress(progress: _SearchProgress | None) -> None:
    if progress is None:
        return
    progress.callback(progress.total_bytes, progress.total_bytes)


def _report_match(
    callback: SearchMatchCallback | None,
    line_number: int,
    line: str,
) -> None:
    if callback is None:
        return
    callback(line_number, line)


def _capture_matching_rows(
    callback: SearchMatchCallback | None,
) -> tuple[list[tuple[int, str]], SearchMatchCallback]:
    rows: list[tuple[int, str]] = []

    def _collector(line_number: int, line: str) -> None:
        rows.append((line_number, line))
        _report_match(callback, line_number, line)

    return rows, _collector


def _dedupe_matching_rows(
    rows: list[tuple[int, str]],
) -> list[tuple[int, str]]:
    seen: set[tuple[int, str]] = set()
    deduped: list[tuple[int, str]] = []
    for entry in rows:
        if entry in seen:
            continue
        seen.add(entry)
        deduped.append(entry)
    return deduped


def search_smtp_conversations(
    log_path: Path,
    term: str,
    *,
    mode: str = MODE_LITERAL,
    fuzzy_threshold: float = DEFAULT_FUZZY_THRESHOLD,
    ignore_case: bool = True,
    materialization: str = MATERIALIZATION_AUTO,
    use_index_cache: bool = False,
    progress_callback: SearchProgressCallback | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> SmtpSearchResult:
    """Return SMTP conversations containing ``term``.

    ``mode`` controls the match syntax. ``literal`` uses exact substring
    matching, ``wildcard`` allows ``*`` and ``?`` wildcards, and ``regex``
    treats ``term`` as a Python regular expression.
    """

    matching_rows, collector = _capture_matching_rows(match_callback)
    result = _search_grouped_entries(
        log_path,
        term,
        _smtp_owner_id,
        mode=mode,
        fuzzy_threshold=fuzzy_threshold,
        ignore_case=ignore_case,
        materialization=materialization,
        use_index_cache=use_index_cache,
        progress_callback=progress_callback,
        match_callback=collector,
        owner_key="smtp",
    )
    result.matching_rows = _dedupe_matching_rows(matching_rows)
    return result


def search_delivery_conversations(
    log_path: Path,
    term: str,
    *,
    mode: str = MODE_LITERAL,
    fuzzy_threshold: float = DEFAULT_FUZZY_THRESHOLD,
    ignore_case: bool = True,
    materialization: str = MATERIALIZATION_AUTO,
    use_index_cache: bool = False,
    progress_callback: SearchProgressCallback | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> SmtpSearchResult:
    """Return delivery conversations containing ``term``."""

    matching_rows, collector = _capture_matching_rows(match_callback)
    result = _search_grouped_entries(
        log_path,
        term,
        _delivery_owner_id,
        mode=mode,
        fuzzy_threshold=fuzzy_threshold,
        ignore_case=ignore_case,
        materialization=materialization,
        use_index_cache=use_index_cache,
        progress_callback=progress_callback,
        match_callback=collector,
        owner_key="delivery",
    )
    result.matching_rows = _dedupe_matching_rows(matching_rows)
    return result


def search_admin_entries(
    log_path: Path,
    term: str,
    *,
    mode: str = MODE_LITERAL,
    fuzzy_threshold: float = DEFAULT_FUZZY_THRESHOLD,
    ignore_case: bool = True,
    materialization: str = MATERIALIZATION_AUTO,
    use_index_cache: bool = False,
    progress_callback: SearchProgressCallback | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> SmtpSearchResult:
    """Return administrative log entries containing ``term``."""

    matching_rows, collector = _capture_matching_rows(match_callback)
    result = _search_grouped_entries(
        log_path,
        term,
        _admin_owner_id,
        mode=mode,
        fuzzy_threshold=fuzzy_threshold,
        ignore_case=ignore_case,
        materialization=materialization,
        use_index_cache=use_index_cache,
        progress_callback=progress_callback,
        match_callback=collector,
        owner_key="admin",
    )
    result.matching_rows = _dedupe_matching_rows(matching_rows)
    return result


def search_imap_retrieval_entries(
    log_path: Path,
    term: str,
    *,
    mode: str = MODE_LITERAL,
    fuzzy_threshold: float = DEFAULT_FUZZY_THRESHOLD,
    ignore_case: bool = True,
    materialization: str = MATERIALIZATION_AUTO,
    use_index_cache: bool = False,
    progress_callback: SearchProgressCallback | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> SmtpSearchResult:
    """Return IMAP retrieval entries containing ``term``."""

    matching_rows, collector = _capture_matching_rows(match_callback)
    result = _search_grouped_entries(
        log_path,
        term,
        _imap_retrieval_owner_id,
        mode=mode,
        fuzzy_threshold=fuzzy_threshold,
        ignore_case=ignore_case,
        materialization=materialization,
        use_index_cache=use_index_cache,
        progress_callback=progress_callback,
        match_callback=collector,
        owner_key="imap-retrieval",
    )
    result.matching_rows = _dedupe_matching_rows(matching_rows)
    return result


def search_ungrouped_entries(
    log_path: Path,
    term: str,
    *,
    mode: str = MODE_LITERAL,
    fuzzy_threshold: float = DEFAULT_FUZZY_THRESHOLD,
    ignore_case: bool = True,
    materialization: str = MATERIALIZATION_AUTO,
    use_index_cache: bool = False,
    progress_callback: SearchProgressCallback | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> SmtpSearchResult:
    """Return ungrouped log entries containing ``term``."""

    matching_rows, collector = _capture_matching_rows(match_callback)
    matcher = _compile_line_matcher(
        term,
        mode,
        ignore_case,
        fuzzy_threshold,
    )
    resolved_materialization = normalize_materialization_mode(materialization)
    expected_passes = 1
    if (
        use_index_cache
        or resolved_materialization != MATERIALIZATION_SINGLE_PASS
    ):
        expected_passes = 2
    progress = _start_search_progress(
        log_path,
        progress_callback,
        expected_passes=expected_passes,
    )
    try:
        if use_index_cache:
            result = _search_ungrouped_with_index(
                log_path,
                term,
                matcher,
                progress=progress,
                match_callback=collector,
            )
            result.matching_rows = _dedupe_matching_rows(matching_rows)
            return result
        if resolved_materialization == MATERIALIZATION_TWO_PASS:
            result = _search_ungrouped_two_pass(
                log_path,
                term,
                matcher,
                progress=progress,
                match_callback=collector,
            )
            result.matching_rows = _dedupe_matching_rows(matching_rows)
            return result

        result = _search_ungrouped_single_pass(
            log_path,
            term,
            matcher,
            auto_fallback=resolved_materialization == MATERIALIZATION_AUTO,
            progress=progress,
            match_callback=collector,
        )
        result.matching_rows = _dedupe_matching_rows(matching_rows)
        return result
    finally:
        _finish_search_progress(progress)


def normalize_materialization_mode(value: str) -> str:
    """Normalize and validate grouped-search materialization strategy."""

    mode = value.strip().lower()
    if mode in SUPPORTED_MATERIALIZATION_MODES:
        return mode
    supported = ", ".join(SUPPORTED_MATERIALIZATION_MODES)
    raise ValueError(
        f"Unsupported materialization mode: {value!r}. "
        f"Expected one of: {supported}."
    )


def _owner_index_cache_key(
    log_path: Path,
    owner_key: str,
) -> tuple[tuple[str, str, int, int], tuple[int, int]]:
    stat = log_path.stat()
    signature = (stat.st_size, stat.st_mtime_ns)
    cache_key = (
        str(log_path.resolve()),
        owner_key,
        signature[0],
        signature[1],
    )
    return cache_key, signature


def _estimate_owner_index_size(
    owner_codes: array,
    owner_ids: list[str],
    owner_first_lines: list[int],
) -> int:
    owner_bytes = sum(len(value) for value in owner_ids)
    return (
        owner_codes.itemsize * len(owner_codes)
        + len(owner_ids) * 64
        + len(owner_first_lines) * 8
        + owner_bytes
    )


def _evict_index_cache_if_needed() -> None:
    global _INDEX_CACHE_BYTES
    while _INDEX_CACHE and _INDEX_CACHE_BYTES > _INDEX_CACHE_MAX_BYTES:
        _, oldest = _INDEX_CACHE.popitem(last=False)
        _INDEX_CACHE_BYTES = max(0, _INDEX_CACHE_BYTES - oldest.size_bytes)


def _cache_owner_line_index(
    cache_key: tuple[str, str, int, int],
    index: _OwnerLineIndex,
) -> None:
    global _INDEX_CACHE_BYTES
    with _INDEX_CACHE_LOCK:
        previous = _INDEX_CACHE.get(cache_key)
        if previous is not None:
            _INDEX_CACHE_BYTES = max(
                0,
                _INDEX_CACHE_BYTES - previous.size_bytes,
            )
        _INDEX_CACHE[cache_key] = index
        _INDEX_CACHE.move_to_end(cache_key)
        _INDEX_CACHE_BYTES += index.size_bytes
        _evict_index_cache_if_needed()


def _lookup_owner_line_index(
    cache_key: tuple[str, str, int, int],
) -> _OwnerLineIndex | None:
    with _INDEX_CACHE_LOCK:
        cached = _INDEX_CACHE.get(cache_key)
        if cached is None:
            return None
        _INDEX_CACHE.move_to_end(cache_key)
        return cached


def _build_grouped_owner_line_index_with_scan(
    log_path: Path,
    owner_for_line: Callable[[str], str | None],
    owner_key: str,
    signature: tuple[int, int],
    matcher: Callable[[str], bool],
    progress: _SearchProgress | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> tuple[_OwnerLineIndex, set[int], list[tuple[int, str]], int]:
    owner_codes = array("i")
    owner_ids: list[str] = []
    owner_first_lines: list[int] = []
    owner_to_code: dict[str, int] = {}
    matched_codes: set[int] = set()
    orphan_matches: list[tuple[int, str]] = []
    current_code = -1
    line_count = 0

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            line_count += 1
            line = raw_line.rstrip("\r\n")
            owner_id = owner_for_line(line)
            if owner_id is not None:
                code = owner_to_code.get(owner_id)
                if code is None:
                    code = len(owner_ids)
                    owner_to_code[owner_id] = code
                    owner_ids.append(owner_id)
                    owner_first_lines.append(line_number)
                current_code = code
                owner_codes.append(code)
                owner_code = code
            else:
                if starts_with_timestamp(line):
                    current_code = -1
                owner_codes.append(current_code)
                owner_code = current_code

            if not matcher(line):
                continue
            _report_match(match_callback, line_number, line)
            if owner_code >= 0:
                matched_codes.add(owner_code)
            else:
                orphan_matches.append((line_number, line))

    size_bytes = _estimate_owner_index_size(
        owner_codes,
        owner_ids,
        owner_first_lines,
    )
    index = _OwnerLineIndex(
        key=owner_key,
        signature=signature,
        owner_codes=owner_codes,
        owner_ids=owner_ids,
        owner_first_lines=owner_first_lines,
        line_count=line_count,
        size_bytes=size_bytes,
    )
    return index, matched_codes, orphan_matches, line_count


def _build_ungrouped_owner_line_index_with_scan(
    log_path: Path,
    signature: tuple[int, int],
    matcher: Callable[[str], bool],
    progress: _SearchProgress | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> tuple[_OwnerLineIndex, set[int], list[tuple[int, str]], int]:
    owner_codes = array("i")
    owner_ids: list[str] = []
    owner_first_lines: list[int] = []
    matched_codes: set[int] = set()
    orphan_matches: list[tuple[int, str]] = []
    current_code = -1
    line_count = 0

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            line_count += 1
            line = raw_line.rstrip("\r\n")
            if starts_with_timestamp(line):
                current_code = len(owner_ids)
                owner_ids.append(f"{line_number}")
                owner_first_lines.append(line_number)
            owner_codes.append(current_code)
            if not matcher(line):
                continue
            _report_match(match_callback, line_number, line)
            if current_code >= 0:
                matched_codes.add(current_code)
            else:
                orphan_matches.append((line_number, line))

    size_bytes = _estimate_owner_index_size(
        owner_codes,
        owner_ids,
        owner_first_lines,
    )
    index = _OwnerLineIndex(
        key="ungrouped",
        signature=signature,
        owner_codes=owner_codes,
        owner_ids=owner_ids,
        owner_first_lines=owner_first_lines,
        line_count=line_count,
        size_bytes=size_bytes,
    )
    return index, matched_codes, orphan_matches, line_count


def _search_grouped_with_index(
    log_path: Path,
    term: str,
    matcher: Callable[[str], bool],
    owner_for_line: Callable[[str], str | None],
    *,
    owner_key: str,
    progress: _SearchProgress | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> SmtpSearchResult:
    cache_key, signature = _owner_index_cache_key(log_path, owner_key)
    cached = _lookup_owner_line_index(cache_key)
    if cached is None:
        (
            index,
            matched_codes,
            orphan_matches,
            total_lines,
        ) = _build_grouped_owner_line_index_with_scan(
            log_path,
            owner_for_line,
            owner_key,
            signature,
            matcher,
            progress=progress,
            match_callback=match_callback,
        )
        _cache_owner_line_index(cache_key, index)
    else:
        index = cached
        matched_codes = set()
        orphan_matches = []
        total_lines = 0
        with log_path.open("r", encoding="utf-8", errors="replace") as handle:
            for zero_based, raw_line in enumerate(handle):
                _advance_search_progress(progress, raw_line)
                total_lines += 1
                line = raw_line.rstrip("\r\n")
                owner_code = (
                    index.owner_codes[zero_based]
                    if zero_based < len(index.owner_codes)
                    else -1
                )
                if not matcher(line):
                    continue
                _report_match(match_callback, zero_based + 1, line)
                if owner_code >= 0:
                    matched_codes.add(owner_code)
                else:
                    orphan_matches.append((zero_based + 1, line))

    if not matched_codes:
        return SmtpSearchResult(
            term=term,
            log_path=log_path,
            conversations=[],
            total_lines=total_lines,
            orphan_matches=orphan_matches,
        )

    conversations_by_code: dict[int, Conversation] = {}
    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for zero_based, raw_line in enumerate(handle):
            _advance_search_progress(progress, raw_line)
            owner_code = (
                index.owner_codes[zero_based]
                if zero_based < len(index.owner_codes)
                else -1
            )
            if owner_code not in matched_codes:
                continue
            conversation = conversations_by_code.get(owner_code)
            if conversation is None:
                conversation = Conversation(
                    message_id=index.owner_ids[owner_code],
                    lines=[],
                    first_line_number=index.owner_first_lines[owner_code],
                )
                conversations_by_code[owner_code] = conversation
            conversation.lines.append(raw_line.rstrip("\r\n"))

    conversations = list(conversations_by_code.values())
    conversations.sort(key=lambda conv: conv.first_line_number)
    return SmtpSearchResult(
        term=term,
        log_path=log_path,
        conversations=conversations,
        total_lines=total_lines,
        orphan_matches=orphan_matches,
    )


def _search_ungrouped_with_index(
    log_path: Path,
    term: str,
    matcher: Callable[[str], bool],
    *,
    progress: _SearchProgress | None = None,
    match_callback: SearchMatchCallback | None = None,
) -> SmtpSearchResult:
    cache_key, signature = _owner_index_cache_key(log_path, "ungrouped")
    cached = _lookup_owner_line_index(cache_key)
    if cached is None:
        (
            index,
            matched_codes,
            orphan_matches,
            total_lines,
        ) = _build_ungrouped_owner_line_index_with_scan(
            log_path,
            signature,
            matcher,
            progress=progress,
            match_callback=match_callback,
        )
        _cache_owner_line_index(cache_key, index)
    else:
        index = cached
        matched_codes = set()
        orphan_matches = []
        total_lines = 0
        with log_path.open("r", encoding="utf-8", errors="replace") as handle:
            for zero_based, raw_line in enumerate(handle):
                _advance_search_progress(progress, raw_line)
                total_lines += 1
                line = raw_line.rstrip("\r\n")
                owner_code = (
                    index.owner_codes[zero_based]
                    if zero_based < len(index.owner_codes)
                    else -1
                )
                if not matcher(line):
                    continue
                _report_match(match_callback, zero_based + 1, line)
                if owner_code >= 0:
                    matched_codes.add(owner_code)
                else:
                    orphan_matches.append((zero_based + 1, line))

    if not matched_codes:
        return SmtpSearchResult(
            term=term,
            log_path=log_path,
            conversations=[],
            total_lines=total_lines,
            orphan_matches=orphan_matches,
        )

    conversations_by_code: dict[int, Conversation] = {}
    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for zero_based, raw_line in enumerate(handle):
            _advance_search_progress(progress, raw_line)
            owner_code = (
                index.owner_codes[zero_based]
                if zero_based < len(index.owner_codes)
                else -1
            )
            if owner_code not in matched_codes:
                continue
            conversation = conversations_by_code.get(owner_code)
            if conversation is None:
                conversation = Conversation(
                    message_id=index.owner_ids[owner_code],
                    lines=[],
                    first_line_number=index.owner_first_lines[owner_code],
                )
                conversations_by_code[owner_code] = conversation
            conversation.lines.append(raw_line.rstrip("\r\n"))

    conversations = list(conversations_by_code.values())
    conversations.sort(key=lambda conv: conv.first_line_number)
    return SmtpSearchResult(
        term=term,
        log_path=log_path,
        conversations=conversations,
        total_lines=total_lines,
        orphan_matches=orphan_matches,
    )


def _search_grouped_entries(
    log_path: Path,
    term: str,
    owner_for_line: Callable[[str], str | None],
    *,
    mode: str,
    fuzzy_threshold: float,
    ignore_case: bool,
    materialization: str,
    use_index_cache: bool,
    progress_callback: SearchProgressCallback | None,
    match_callback: SearchMatchCallback | None,
    owner_key: str,
) -> SmtpSearchResult:
    matcher = _compile_line_matcher(
        term,
        mode,
        ignore_case,
        fuzzy_threshold,
    )
    resolved_materialization = normalize_materialization_mode(materialization)
    expected_passes = 1
    if (
        use_index_cache
        or resolved_materialization != MATERIALIZATION_SINGLE_PASS
    ):
        expected_passes = 2
    progress = _start_search_progress(
        log_path,
        progress_callback,
        expected_passes=expected_passes,
    )
    try:
        if use_index_cache:
            return _search_grouped_with_index(
                log_path,
                term,
                matcher,
                owner_for_line,
                owner_key=owner_key,
                progress=progress,
                match_callback=match_callback,
            )
        if resolved_materialization == MATERIALIZATION_TWO_PASS:
            return _search_grouped_two_pass(
                log_path,
                term,
                matcher,
                owner_for_line,
                progress=progress,
                match_callback=match_callback,
            )
        return _search_grouped_single_pass(
            log_path,
            term,
            matcher,
            owner_for_line,
            auto_fallback=resolved_materialization == MATERIALIZATION_AUTO,
            progress=progress,
            match_callback=match_callback,
        )
    finally:
        _finish_search_progress(progress)


def _search_grouped_two_pass(
    log_path: Path,
    term: str,
    matcher: Callable[[str], bool],
    owner_for_line: Callable[[str], str | None],
    *,
    progress: _SearchProgress | None,
    match_callback: SearchMatchCallback | None,
) -> SmtpSearchResult:
    matched_ids, orphan_matches, total_lines = _scan_grouped_matches(
        log_path,
        matcher,
        owner_for_line,
        progress=progress,
        match_callback=match_callback,
    )
    if not matched_ids:
        return SmtpSearchResult(
            term=term,
            log_path=log_path,
            conversations=[],
            total_lines=total_lines,
            orphan_matches=orphan_matches,
        )

    conversations = _collect_grouped_conversations(
        log_path,
        matched_ids,
        owner_for_line,
        progress=progress,
    )
    return SmtpSearchResult(
        term=term,
        log_path=log_path,
        conversations=conversations,
        total_lines=total_lines,
        orphan_matches=orphan_matches,
    )


def _scan_grouped_matches(
    log_path: Path,
    matcher: Callable[[str], bool],
    owner_for_line: Callable[[str], str | None],
    *,
    progress: _SearchProgress | None,
    match_callback: SearchMatchCallback | None,
) -> tuple[set[str], list[tuple[int, str]], int]:
    matched_ids: set[str] = set()
    orphan_matches: list[tuple[int, str]] = []
    total_lines = 0
    current_id: str | None = None

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            total_lines += 1
            line = raw_line.rstrip("\r\n")
            line_owner_id: str | None = None
            owner_id = owner_for_line(line)
            if owner_id is not None:
                current_id = owner_id
                line_owner_id = owner_id
            elif starts_with_timestamp(line):
                current_id = None
            elif current_id is not None:
                line_owner_id = current_id

            if matcher(line):
                _report_match(match_callback, line_number, line)
                if line_owner_id is not None:
                    matched_ids.add(line_owner_id)
                else:
                    orphan_matches.append((line_number, line))
    return matched_ids, orphan_matches, total_lines


def _collect_grouped_conversations(
    log_path: Path,
    matched_ids: set[str],
    owner_for_line: Callable[[str], str | None],
    *,
    progress: _SearchProgress | None,
) -> list[Conversation]:
    builders: dict[str, Conversation] = {}
    current_id: str | None = None

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            line = raw_line.rstrip("\r\n")
            line_owner_id: str | None = None
            owner_id = owner_for_line(line)
            if owner_id is not None:
                current_id = owner_id
                line_owner_id = owner_id
            elif starts_with_timestamp(line):
                current_id = None
            elif current_id is not None:
                line_owner_id = current_id

            if line_owner_id is None or line_owner_id not in matched_ids:
                continue
            conversation = builders.get(line_owner_id)
            if conversation is None:
                conversation = Conversation(
                    message_id=line_owner_id,
                    lines=[],
                    first_line_number=line_number,
                )
                builders[line_owner_id] = conversation
            conversation.lines.append(line)

    conversations = list(builders.values())
    conversations.sort(key=lambda conv: conv.first_line_number)
    return conversations


def _search_grouped_single_pass(
    log_path: Path,
    term: str,
    matcher: Callable[[str], bool],
    owner_for_line: Callable[[str], str | None],
    *,
    auto_fallback: bool,
    progress: _SearchProgress | None,
    match_callback: SearchMatchCallback | None,
) -> SmtpSearchResult:
    builders: dict[str, Conversation] = {}
    matched_ids: set[str] = set()
    orphan_matches: list[tuple[int, str]] = []
    total_lines = 0
    current_id: str | None = None
    sample_match_lines = 0
    sample_owner_ids: set[str] = set()
    sample_matched_owner_ids: set[str] = set()

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            total_lines += 1
            line = raw_line.rstrip("\r\n")
            line_owner_id: str | None = None
            owner_id = owner_for_line(line)
            if owner_id is not None:
                current_id = owner_id
                line_owner_id = owner_id
                conversation = builders.get(owner_id)
                if conversation is None:
                    conversation = Conversation(
                        message_id=owner_id,
                        lines=[],
                        first_line_number=line_number,
                    )
                    builders[owner_id] = conversation
                conversation.lines.append(line)
            elif starts_with_timestamp(line):
                current_id = None
            elif current_id is not None:
                line_owner_id = current_id
                current_conversation = builders.get(current_id)
                if current_conversation is None:
                    current_conversation = Conversation(
                        message_id=current_id,
                        lines=[],
                        first_line_number=line_number,
                    )
                    builders[current_id] = current_conversation
                current_conversation.lines.append(line)

            is_match = matcher(line)
            if is_match:
                _report_match(match_callback, line_number, line)
                if line_owner_id is not None:
                    matched_ids.add(line_owner_id)
                else:
                    orphan_matches.append((line_number, line))

            if line_number > _AUTO_SAMPLE_LINES:
                continue
            if owner_id is not None:
                sample_owner_ids.add(owner_id)
            if is_match:
                sample_match_lines += 1
                if line_owner_id is not None:
                    sample_matched_owner_ids.add(line_owner_id)
            if (
                auto_fallback
                and line_number == _AUTO_SAMPLE_LINES
                and _looks_like_sparse_query(
                    line_number,
                    sample_match_lines,
                    len(sample_owner_ids),
                    len(sample_matched_owner_ids),
                )
            ):
                return _search_grouped_two_pass(
                    log_path,
                    term,
                    matcher,
                    owner_for_line,
                    progress=progress,
                    match_callback=match_callback,
                )

    conversations = [
        builders[mid]
        for mid in matched_ids
        if mid in builders
    ]
    conversations.sort(key=lambda conv: conv.first_line_number)
    return SmtpSearchResult(
        term=term,
        log_path=log_path,
        conversations=conversations,
        total_lines=total_lines,
        orphan_matches=orphan_matches,
    )


def _looks_like_sparse_query(
    sample_lines: int,
    sample_match_lines: int,
    sample_owner_count: int,
    sample_matched_owner_count: int,
) -> bool:
    if sample_lines <= 0:
        return False
    line_match_ratio = sample_match_lines / sample_lines
    if sample_owner_count <= 0:
        return line_match_ratio <= _AUTO_MAX_LINE_MATCH_RATIO
    owner_match_ratio = sample_matched_owner_count / sample_owner_count
    return (
        line_match_ratio <= _AUTO_MAX_LINE_MATCH_RATIO
        and owner_match_ratio <= _AUTO_MAX_OWNER_MATCH_RATIO
    )


def _search_ungrouped_two_pass(
    log_path: Path,
    term: str,
    matcher: Callable[[str], bool],
    *,
    progress: _SearchProgress | None,
    match_callback: SearchMatchCallback | None,
) -> SmtpSearchResult:
    matched_ids, orphan_matches, total_lines = _scan_ungrouped_matches(
        log_path,
        matcher,
        progress=progress,
        match_callback=match_callback,
    )
    if not matched_ids:
        return SmtpSearchResult(
            term=term,
            log_path=log_path,
            conversations=[],
            total_lines=total_lines,
            orphan_matches=orphan_matches,
        )

    conversations = _collect_ungrouped_conversations(
        log_path,
        matched_ids,
        progress=progress,
    )
    return SmtpSearchResult(
        term=term,
        log_path=log_path,
        conversations=conversations,
        total_lines=total_lines,
        orphan_matches=orphan_matches,
    )


def _scan_ungrouped_matches(
    log_path: Path,
    matcher: Callable[[str], bool],
    *,
    progress: _SearchProgress | None,
    match_callback: SearchMatchCallback | None,
) -> tuple[set[str], list[tuple[int, str]], int]:
    matched_ids: set[str] = set()
    orphan_matches: list[tuple[int, str]] = []
    total_lines = 0
    current_id: str | None = None

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            total_lines += 1
            line = raw_line.rstrip("\r\n")
            line_owner_id: str | None = None
            if starts_with_timestamp(line):
                current_id = f"{line_number}"
                line_owner_id = current_id
            elif current_id is not None:
                line_owner_id = current_id

            if matcher(line):
                _report_match(match_callback, line_number, line)
                if line_owner_id is not None:
                    matched_ids.add(line_owner_id)
                else:
                    orphan_matches.append((line_number, line))
    return matched_ids, orphan_matches, total_lines


def _collect_ungrouped_conversations(
    log_path: Path,
    matched_ids: set[str],
    *,
    progress: _SearchProgress | None,
) -> list[Conversation]:
    builders: dict[str, Conversation] = {}
    current_id: str | None = None

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            line = raw_line.rstrip("\r\n")
            line_owner_id: str | None = None
            if starts_with_timestamp(line):
                current_id = f"{line_number}"
                line_owner_id = current_id
            elif current_id is not None:
                line_owner_id = current_id

            if line_owner_id is None or line_owner_id not in matched_ids:
                continue
            conversation = builders.get(line_owner_id)
            if conversation is None:
                conversation = Conversation(
                    message_id=line_owner_id,
                    lines=[],
                    first_line_number=line_number,
                )
                builders[line_owner_id] = conversation
            conversation.lines.append(line)

    conversations = list(builders.values())
    conversations.sort(key=lambda conv: conv.first_line_number)
    return conversations


def _search_ungrouped_single_pass(
    log_path: Path,
    term: str,
    matcher: Callable[[str], bool],
    *,
    auto_fallback: bool,
    progress: _SearchProgress | None,
    match_callback: SearchMatchCallback | None,
) -> SmtpSearchResult:
    builders: dict[str, Conversation] = {}
    matched_ids: set[str] = set()
    orphan_matches: list[tuple[int, str]] = []
    total_lines = 0
    current_id: str | None = None
    sample_match_lines = 0
    sample_owner_count = 0
    sample_matched_owner_ids: set[str] = set()

    with log_path.open("r", encoding="utf-8", errors="replace") as handle:
        for line_number, raw_line in enumerate(handle, start=1):
            _advance_search_progress(progress, raw_line)
            total_lines += 1
            line = raw_line.rstrip("\r\n")
            line_owner_id: str | None = None
            owner_id: str | None = None
            if starts_with_timestamp(line):
                owner_id = f"{line_number}"
                current_id = owner_id
                line_owner_id = owner_id
                sample_owner_count += 1
                conversation = Conversation(
                    message_id=owner_id,
                    lines=[],
                    first_line_number=line_number,
                )
                builders[owner_id] = conversation
                conversation.lines.append(line)
            elif current_id is not None:
                line_owner_id = current_id
                current_conversation = builders.get(current_id)
                if current_conversation is None:
                    current_conversation = Conversation(
                        message_id=current_id,
                        lines=[],
                        first_line_number=line_number,
                    )
                    builders[current_id] = current_conversation
                current_conversation.lines.append(line)

            is_match = matcher(line)
            if is_match:
                _report_match(match_callback, line_number, line)
                if line_owner_id is not None:
                    matched_ids.add(line_owner_id)
                else:
                    orphan_matches.append((line_number, line))

            if line_number > _AUTO_SAMPLE_LINES:
                continue
            if is_match:
                sample_match_lines += 1
                if line_owner_id is not None:
                    sample_matched_owner_ids.add(line_owner_id)
            if (
                auto_fallback
                and line_number == _AUTO_SAMPLE_LINES
                and _looks_like_sparse_query(
                    line_number,
                    sample_match_lines,
                    sample_owner_count,
                    len(sample_matched_owner_ids),
                )
            ):
                return _search_ungrouped_two_pass(
                    log_path,
                    term,
                    matcher,
                    progress=progress,
                    match_callback=match_callback,
                )

    conversations = [
        builders[mid]
        for mid in matched_ids
        if mid in builders
    ]
    conversations.sort(key=lambda conv: conv.first_line_number)
    return SmtpSearchResult(
        term=term,
        log_path=log_path,
        conversations=conversations,
        total_lines=total_lines,
        orphan_matches=orphan_matches,
    )


def _compile_match_pattern(
    term: str,
    mode: str,
    ignore_case: bool,
) -> re.Pattern[str]:
    flags = re.IGNORECASE if ignore_case else 0
    resolved_mode = normalize_search_mode(mode)

    if resolved_mode == MODE_LITERAL:
        source = re.escape(term)
    elif resolved_mode == MODE_WILDCARD:
        source = wildcard_to_regex(term)
    elif resolved_mode == MODE_REGEX:
        source = term
    else:  # pragma: no cover - normalize_search_mode gates this
        raise ValueError(f"Unsupported search mode: {mode!r}")

    try:
        return re.compile(source, flags)
    except re.error as exc:
        if resolved_mode == MODE_REGEX:
            raise ValueError(f"Invalid regex pattern: {exc}") from exc
        raise


def _compile_line_matcher(
    term: str,
    mode: str,
    ignore_case: bool,
    fuzzy_threshold: float,
) -> Callable[[str], bool]:
    resolved_mode = normalize_search_mode(mode)
    if resolved_mode == MODE_LITERAL:
        return _compile_literal_line_matcher(term, ignore_case)
    if resolved_mode == MODE_WILDCARD:
        return _compile_wildcard_line_matcher(term, ignore_case)
    if resolved_mode == MODE_REGEX:
        return _compile_regex_line_matcher(term, ignore_case)

    if resolved_mode == MODE_FUZZY:
        threshold = normalize_fuzzy_threshold(fuzzy_threshold)
        normalized_term = term.lower() if ignore_case else term
        return _compile_fuzzy_line_matcher(
            normalized_term,
            threshold,
            ignore_case=ignore_case,
        )

    pattern = _compile_match_pattern(term, resolved_mode, ignore_case)
    search = pattern.search
    return lambda line: search(line) is not None


def _compile_literal_line_matcher(
    term: str,
    ignore_case: bool,
) -> Callable[[str], bool]:
    if ignore_case:
        term = term.lower()
        lower = str.lower
        return lambda line: term in lower(line)
    return lambda line: term in line


def _compile_wildcard_line_matcher(
    term: str,
    ignore_case: bool,
) -> Callable[[str], bool]:
    if "*" not in term and "?" not in term:
        return _compile_literal_line_matcher(term, ignore_case)
    pattern = _compile_match_pattern(term, MODE_WILDCARD, ignore_case)
    search = pattern.search
    literal_hint = _longest_wildcard_literal(term)
    if len(literal_hint) < 2:
        return lambda line: search(line) is not None
    if ignore_case:
        hint = literal_hint.lower()
        return lambda line: hint in line.lower() and search(line) is not None
    return lambda line: literal_hint in line and search(line) is not None


def _longest_wildcard_literal(term: str) -> str:
    longest = ""
    current: list[str] = []
    for char in term:
        if char in {"*", "?"}:
            token = "".join(current)
            if len(token) > len(longest):
                longest = token
            current.clear()
            continue
        current.append(char)
    token = "".join(current)
    if len(token) > len(longest):
        longest = token
    return longest


def _compile_regex_line_matcher(
    term: str,
    ignore_case: bool,
) -> Callable[[str], bool]:
    if _is_plain_regex_literal(term):
        return _compile_literal_line_matcher(term, ignore_case)
    pattern = _compile_match_pattern(term, MODE_REGEX, ignore_case)
    search = pattern.search
    return lambda line: search(line) is not None


def _is_plain_regex_literal(term: str) -> bool:
    regex_meta = {
        "\\",
        ".",
        "^",
        "$",
        "*",
        "+",
        "?",
        "{",
        "}",
        "[",
        "]",
        "|",
        "(",
        ")",
    }
    return all(char not in regex_meta for char in term)


def _compile_fuzzy_line_matcher(
    term: str,
    threshold: float,
    *,
    ignore_case: bool,
) -> Callable[[str], bool]:
    if _rapidfuzz_fuzz is not None:
        return _compile_rapidfuzz_line_matcher(
            term,
            threshold,
            ignore_case=ignore_case,
        )
    term_len = len(term)
    if term_len <= 0:
        return lambda _line: False
    stride = max(1, term_len // _FUZZY_STRIDE_DIVISOR)
    anchor_offsets = _build_anchor_offsets(term_len)
    anchor_chars = tuple(term[offset] for offset in anchor_offsets)

    if ignore_case:
        return lambda line: _fuzzy_line_match(
            term,
            line.lower(),
            threshold,
            term_len,
            stride,
            anchor_offsets,
            anchor_chars,
        )
    return lambda line: _fuzzy_line_match(
        term,
        line,
        threshold,
        term_len,
        stride,
        anchor_offsets,
        anchor_chars,
    )


def _compile_rapidfuzz_line_matcher(
    term: str,
    threshold: float,
    *,
    ignore_case: bool,
) -> Callable[[str], bool]:
    term_len = len(term)
    if term_len <= 0:
        return lambda _line: False
    cutoff = threshold * 100.0
    if ignore_case:
        return lambda line: _rapidfuzz_line_match(
            term,
            line.lower(),
            cutoff,
        )
    return lambda line: _rapidfuzz_line_match(term, line, cutoff)


def _rapidfuzz_line_match(
    term: str,
    line: str,
    cutoff: float,
) -> bool:
    if not term:
        return False
    if term in line:
        return True
    score = _rapidfuzz_fuzz.partial_ratio(  # type: ignore[union-attr]
        term,
        line,
        score_cutoff=cutoff,
    )
    return score >= cutoff


def _fuzzy_line_match(
    term: str,
    line: str,
    threshold: float,
    term_len: int,
    stride: int,
    anchor_offsets: tuple[int, ...],
    anchor_chars: tuple[str, ...],
) -> bool:
    if not term:
        return False
    if term in line:
        return True
    if not line:
        return False
    if len(line) <= term_len:
        return SequenceMatcher(None, term, line).ratio() >= threshold

    max_start = len(line) - term_len
    starts = _candidate_window_starts(
        line,
        max_start,
        stride,
        anchor_offsets,
        anchor_chars,
    )
    matcher = SequenceMatcher(None, term)
    best_start = 0
    best_ratio = 0.0

    for start in starts:
        ratio = _window_similarity(matcher, line, start, term_len, threshold)
        if ratio >= threshold:
            return True
        if ratio > best_ratio:
            best_ratio = ratio
            best_start = start

    if best_ratio <= 0.0 or stride <= 1:
        return False

    begin = max(0, best_start - stride + 1)
    end = min(max_start, best_start + stride - 1)
    for start in range(begin, end + 1):
        if start in starts:
            continue
        ratio = _window_similarity(matcher, line, start, term_len, threshold)
        if ratio >= threshold:
            return True
    return False


def _build_anchor_offsets(term_len: int) -> tuple[int, ...]:
    if term_len <= 1:
        return (0,)
    midpoint = term_len // 2
    return (0, midpoint, term_len - 1)


def _candidate_window_starts(
    line: str,
    max_start: int,
    stride: int,
    anchor_offsets: tuple[int, ...],
    anchor_chars: tuple[str, ...],
) -> set[int]:
    starts: set[int] = {0, max_start}
    for start in range(0, max_start + 1, stride):
        starts.add(start)

    for offset, char in zip(anchor_offsets, anchor_chars):
        hits = 0
        position = line.find(char)
        while position != -1 and hits < _FUZZY_ANCHOR_LIMIT:
            start = position - offset
            if start < 0:
                start = 0
            if start > max_start:
                start = max_start
            starts.add(start)
            position = line.find(char, position + 1)
            hits += 1
    return starts


def _window_similarity(
    matcher: SequenceMatcher,
    line: str,
    start: int,
    window: int,
    threshold: float,
) -> float:
    chunk = line[start:start + window]
    matcher.set_seq2(chunk)
    if matcher.real_quick_ratio() < threshold:
        return 0.0
    if matcher.quick_ratio() < threshold:
        return 0.0
    return matcher.ratio()


def _owner_strategy_for_kind(
    kind: str,
) -> tuple[str, Callable[[str], str | None] | None]:
    kind_key = normalize_kind(kind)
    if kind_key in {KIND_SMTP, KIND_IMAP, KIND_POP}:
        return "smtp", _smtp_owner_id
    if kind_key == KIND_DELIVERY:
        return "delivery", _delivery_owner_id
    if kind_key == KIND_ADMINISTRATIVE:
        return "admin", _admin_owner_id
    if kind_key == KIND_IMAP_RETRIEVAL:
        return "imap-retrieval", _imap_retrieval_owner_id
    if is_search_ungrouped_kind(kind_key):
        return "ungrouped", None
    raise ValueError(f"Unsupported log kind: {kind}")


def has_search_index(log_path: Path, kind: str) -> bool:
    """Return whether an owner index is cached.

    The cache key is scoped by ``log_path`` and ``kind``.
    """

    owner_key, _owner_for_line = _owner_strategy_for_kind(kind)
    cache_key, _signature = _owner_index_cache_key(log_path, owner_key)
    return _lookup_owner_line_index(cache_key) is not None


def prime_search_index(log_path: Path, kind: str) -> None:
    """Build and cache a search owner index for ``log_path`` and ``kind``."""

    owner_key, owner_for_line = _owner_strategy_for_kind(kind)
    cache_key, signature = _owner_index_cache_key(log_path, owner_key)
    if _lookup_owner_line_index(cache_key) is not None:
        return

    never_match = lambda _line: False
    if owner_key == "ungrouped":
        index, _matched, _orphans, _count = (
            _build_ungrouped_owner_line_index_with_scan(
                log_path,
                signature,
                never_match,
            )
        )
    else:
        assert owner_for_line is not None
        index, _matched, _orphans, _count = (
            _build_grouped_owner_line_index_with_scan(
                log_path,
                owner_for_line,
                owner_key,
                signature,
                never_match,
            )
        )
    _cache_owner_line_index(cache_key, index)


def _smtp_owner_id(line: str) -> str | None:
    match = _SMTP_OWNER_PATTERN.match(line)
    if match is None:
        return None
    return match.group(1)


def _delivery_owner_id(line: str) -> str | None:
    match = _DELIVERY_OWNER_PATTERN.match(line)
    if match is None:
        return None
    return match.group(1)


def _imap_retrieval_owner_id(line: str) -> str | None:
    match = _IMAP_RETRIEVAL_OWNER_PATTERN.match(line)
    if match is None:
        return None
    return match.group(1)


def _admin_owner_id(line: str) -> str | None:
    match = _ADMIN_OWNER_PATTERN.match(line)
    if match is not None:
        return f"{match.group('ip')} {match.group('time')}"
    trailing_match = _ADMIN_TRAILING_OWNER_PATTERN.match(line)
    if trailing_match is None:
        return None
    return f"{trailing_match.group('ip')} {trailing_match.group('time')}"


def get_search_function(
    kind: str,
) -> SearchFunction | None:
    """Return the search function for the given log kind."""

    kind_key = normalize_kind(kind)
    if kind_key == KIND_SMTP:
        return search_smtp_conversations
    if kind_key == KIND_IMAP:
        return search_smtp_conversations
    if kind_key == KIND_POP:
        return search_smtp_conversations
    if kind_key == KIND_DELIVERY:
        return search_delivery_conversations
    if kind_key == KIND_ADMINISTRATIVE:
        return search_admin_entries
    if kind_key == KIND_IMAP_RETRIEVAL:
        return search_imap_retrieval_entries
    if is_search_ungrouped_kind(kind_key):
        return search_ungrouped_entries
    return None



from sm_logtool.syntax import (
    TOKEN_EMAIL,
    TOKEN_IP,
    TOKEN_LINE_NUMBER,
    TOKEN_PROTO_IMAP,
    TOKEN_RESPONSE,
    TOKEN_STATUS_BAD,
    TOKEN_TIMESTAMP,
    spans_for_line,
)


def _has_span(spans, token, start, end):
    for span in spans:
        if span.token != token:
            continue
        if span.start <= start and span.end >= end:
            return True
    return False


def test_highlight_smtp_line_styles_timestamp_and_ip():
    line = (
        "23:59:56.065 [111.70.33.193][39603817] cmd: "
        "EHLO example.com"
    )
    spans = spans_for_line("smtp", line)
    assert _has_span(spans, TOKEN_TIMESTAMP, 0, len("23:59:56.065"))
    ip_start = line.index("111.70.33.193")
    ip_end = ip_start + len("111.70.33.193")
    assert _has_span(spans, TOKEN_IP, ip_start, ip_end)


def test_highlight_delivery_line_styles_email():
    line = (
        "23:59:59.117 [72495970] Starting local delivery to "
        "andy@shasta.com"
    )
    spans = spans_for_line("delivery", line)
    email_start = line.index("andy@shasta.com")
    email_end = email_start + len("andy@shasta.com")
    assert _has_span(spans, TOKEN_EMAIL, email_start, email_end)


def test_highlight_orphan_prefix_dim():
    line = (
        "123: 00:00:01 [1.1.1.1][ABC] Connection initiated"
    )
    spans = spans_for_line("smtp", line)
    prefix_end = len("123: ")
    assert _has_span(spans, TOKEN_LINE_NUMBER, 0, prefix_end)


def test_highlight_admin_protocol_and_status():
    line = (
        "23:59:57.727 [178.216.28.19] IMAP Login failed: "
        "User [204be204] not found"
    )
    spans = spans_for_line("administrative", line)
    proto_start = line.index("IMAP")
    proto_end = proto_start + len("IMAP")
    assert _has_span(spans, TOKEN_PROTO_IMAP, proto_start, proto_end)
    status_start = line.index("failed")
    status_end = status_start + len("failed")
    assert _has_span(spans, TOKEN_STATUS_BAD, status_start, status_end)


def test_response_code_not_highlighted_in_timestamp_columns():
    line = (
        "[2026.02.18] 05:01:30.507 [198.51.100.23][30216663] "
        "rsp: 334 VXNlcm5hbWU6"
    )
    spans = spans_for_line("smtp", line)

    ts_code_start = line.index("507")
    ts_code_end = ts_code_start + len("507")
    assert not _has_span(spans, TOKEN_RESPONSE, ts_code_start, ts_code_end)

    rsp_code_start = line.index("334")
    rsp_code_end = rsp_code_start + len("334")
    assert _has_span(spans, TOKEN_RESPONSE, rsp_code_start, rsp_code_end)


def test_blocked_sender_checks_not_highlighted_as_bad_status():
    line = (
        "00:05:58.836 [84012980] Blocked Sender Checks started."
    )
    spans = spans_for_line("delivery", line)
    blocked_start = line.index("Blocked")
    blocked_end = blocked_start + len("Blocked")
    assert not _has_span(
        spans,
        TOKEN_STATUS_BAD,
        blocked_start,
        blocked_end,
    )


def test_failed_false_status_field_not_highlighted_as_bad_status():
    line = (
        "00:06:07.424 [84012980] Removing Spool message: "
        "Killed: False, Failed: False, Finished: True"
    )
    spans = spans_for_line("delivery", line)
    failed_start = line.index("Failed")
    failed_end = failed_start + len("Failed")
    assert not _has_span(
        spans,
        TOKEN_STATUS_BAD,
        failed_start,
        failed_end,
    )


def test_block_outcome_markers_are_highlighted_as_bad_status():
    action_line = (
        "00:06:03.076 [84012980] Blocking sender <sender@example.test> "
        "for <recipient@example.test>. Action: MoveToJunk"
    )
    action_spans = spans_for_line("delivery", action_line)
    action_start = action_line.index("Action: MoveToJunk")
    action_end = action_start + len("Action: MoveToJunk")
    assert _has_span(
        action_spans,
        TOKEN_STATUS_BAD,
        action_start,
        action_end,
    )

    reason_line = (
        "00:06:03.077 [84012980] REASON: User "
        "'<recipient@example.test>' has <sender@example.test> "
        "on their blocked list."
    )
    reason_spans = spans_for_line("delivery", reason_line)
    reason_start = reason_line.index("on their blocked list")
    reason_end = reason_start + len("on their blocked list")
    assert _has_span(
        reason_spans,
        TOKEN_STATUS_BAD,
        reason_start,
        reason_end,
    )

    country_line = (
        "00:06:04.000 [84012980] REASON: Sender domain "
        "<sender@example.test> is on the blocked country list."
    )
    country_spans = spans_for_line("delivery", country_line)
    country_start = country_line.index("on the blocked country list")
    country_end = country_start + len("on the blocked country list")
    assert _has_span(
        country_spans,
        TOKEN_STATUS_BAD,
        country_start,
        country_end,
    )



"""Unittest bridge that executes the pytest suite."""

from __future__ import annotations

from pathlib import Path
import subprocess
import sys
import unittest


def _running_under_pytest() -> bool:
    argv = " ".join(sys.argv).lower()
    return "pytest" in argv


@unittest.skipIf(
    _running_under_pytest(),
    "Bridge test is only used by unittest discover.",
)
class TestPytestBridge(unittest.TestCase):
    """Run pytest from unittest discover to keep both entrypoints valid."""

    def test_pytest_suite_passes(self) -> None:
        repo_root = Path(__file__).resolve().parent.parent
        bridge_path = Path(__file__).resolve()
        command = [
            sys.executable,
            "-m",
            "pytest",
            "-q",
            "--ignore",
            str(bridge_path),
        ]
        completed = subprocess.run(
            command,
            cwd=repo_root,
            capture_output=True,
            text=True,
            check=False,
        )
        if completed.returncode == 0:
            return

        details = "\n".join(
            [
                "pytest bridge failed.",
                f"Command: {' '.join(command)}",
                f"Exit code: {completed.returncode}",
                f"stdout:\n{completed.stdout}",
                f"stderr:\n{completed.stderr}",
            ]
        )
        self.fail(details)



"""Shared syntax highlighting styles for CLI and TUI output."""

from __future__ import annotations

from rich.style import Style
from rich.text import Text

from .syntax import (
    TOKEN_BRACKET,
    TOKEN_COMMAND,
    TOKEN_EMAIL,
    TOKEN_HEADER,
    TOKEN_ID,
    TOKEN_IP,
    TOKEN_LINE_NUMBER,
    TOKEN_MESSAGE_ID,
    TOKEN_PROTO_ACTIVESYNC,
    TOKEN_PROTO_API,
    TOKEN_PROTO_CALDAV,
    TOKEN_PROTO_CARDDAV,
    TOKEN_PROTO_EAS,
    TOKEN_PROTO_IMAP,
    TOKEN_PROTO_POP,
    TOKEN_PROTO_SMTP,
    TOKEN_PROTO_USER,
    TOKEN_PROTO_WEBMAIL,
    TOKEN_PROTO_XMPP,
    TOKEN_RESPONSE,
    TOKEN_SECTION,
    TOKEN_STATUS_BAD,
    TOKEN_STATUS_GOOD,
    TOKEN_SUMMARY,
    TOKEN_TAG,
    TOKEN_TERM,
    TOKEN_TIMESTAMP,
    spans_for_line,
)

TOKEN_STYLES: dict[str, Style] = {
    TOKEN_HEADER: Style(bold=True),
    TOKEN_SECTION: Style(bold=True),
    TOKEN_SUMMARY: Style(bold=True),
    TOKEN_TERM: Style(color="magenta", bold=True),
    TOKEN_TIMESTAMP: Style(color="cyan", bold=True),
    TOKEN_BRACKET: Style(dim=True),
    TOKEN_IP: Style(color="bright_blue"),
    TOKEN_ID: Style(color="magenta"),
    TOKEN_TAG: Style(color="cyan"),
    TOKEN_EMAIL: Style(color="bright_magenta"),
    TOKEN_COMMAND: Style(color="green"),
    TOKEN_RESPONSE: Style(color="yellow"),
    TOKEN_LINE_NUMBER: Style(dim=True),
    TOKEN_MESSAGE_ID: Style(color="bright_cyan"),
    TOKEN_STATUS_BAD: Style(color="bright_red", bold=True),
    TOKEN_STATUS_GOOD: Style(color="bright_green", bold=True),
    TOKEN_PROTO_SMTP: Style(color="#7bd88f", bold=True),
    TOKEN_PROTO_IMAP: Style(color="#4aa3ff", bold=True),
    TOKEN_PROTO_POP: Style(color="#ffcc66", bold=True),
    TOKEN_PROTO_USER: Style(color="#ff8ec7", bold=True),
    TOKEN_PROTO_WEBMAIL: Style(color="#6be0ff", bold=True),
    TOKEN_PROTO_ACTIVESYNC: Style(color="#ff6b6b", bold=True),
    TOKEN_PROTO_EAS: Style(color="#ffd166", bold=True),
    TOKEN_PROTO_CALDAV: Style(color="#8b7bff", bold=True),
    TOKEN_PROTO_CARDDAV: Style(color="#b388ff", bold=True),
    TOKEN_PROTO_XMPP: Style(color="#4ef0b7", bold=True),
    TOKEN_PROTO_API: Style(color="#ff9f1c", bold=True),
}


def highlight_line(kind: str, line: str) -> Text:
    """Return a Rich ``Text`` line with SmarterMail syntax highlighting."""

    text = Text(line)
    limit = len(line)
    for span in spans_for_line(kind, line):
        style = TOKEN_STYLES.get(span.token)
        if style is None:
            continue
        start = _clamp(span.start, limit)
        end = _clamp(span.end, limit)
        if start >= end:
            continue
        text.stylize(style, start, end)
    return text


def _clamp(value: int, upper: int) -> int:
    if value < 0:
        return 0
    if value > upper:
        return upper
    return value



"""Search mode definitions shared by CLI and TUI."""

from __future__ import annotations

import re

MODE_LITERAL = "literal"
MODE_WILDCARD = "wildcard"
MODE_REGEX = "regex"
MODE_FUZZY = "fuzzy"

DEFAULT_FUZZY_THRESHOLD = 0.75
MIN_FUZZY_THRESHOLD = 0.0
MAX_FUZZY_THRESHOLD = 1.0

SUPPORTED_SEARCH_MODES = (
    MODE_LITERAL,
    MODE_WILDCARD,
    MODE_REGEX,
    MODE_FUZZY,
)

SEARCH_MODE_LABELS = {
    MODE_LITERAL: "Literal",
    MODE_WILDCARD: "Wildcard",
    MODE_REGEX: "Regex",
    MODE_FUZZY: "Fuzzy",
}

SEARCH_MODE_DESCRIPTIONS = {
    MODE_LITERAL: "Exact substring match.",
    MODE_WILDCARD: "'*' matches many chars, '?' matches one char.",
    MODE_REGEX: "Python re syntax (PCRE-like, not full PCRE).",
    MODE_FUZZY: "Approximate line match using similarity threshold.",
}


def normalize_search_mode(value: str | None) -> str:
    """Validate and normalize a search mode value."""

    mode = (value or MODE_LITERAL).strip().lower()
    if mode in SUPPORTED_SEARCH_MODES:
        return mode

    choices = ", ".join(SUPPORTED_SEARCH_MODES)
    raise ValueError(
        f"Unsupported search mode: {value!r}. Choose one of: {choices}."
    )


def normalize_fuzzy_threshold(value: float | None) -> float:
    """Validate and normalize fuzzy threshold."""

    if value is None:
        return DEFAULT_FUZZY_THRESHOLD

    threshold = float(value)
    if MIN_FUZZY_THRESHOLD <= threshold <= MAX_FUZZY_THRESHOLD:
        return threshold

    raise ValueError(
        "Invalid fuzzy threshold: "
        f"{value!r}. Choose a value between "
        f"{MIN_FUZZY_THRESHOLD:.2f} and {MAX_FUZZY_THRESHOLD:.2f}."
    )


def wildcard_to_regex(pattern: str) -> str:
    """Convert wildcard text using ``*`` and ``?`` to regex source."""

    parts: list[str] = []
    for char in pattern:
        if char == "*":
            parts.append(".*")
            continue
        if char == "?":
            parts.append(".")
            continue
        parts.append(re.escape(char))
    return "".join(parts)



from sm_logtool import log_kinds


def test_search_ungrouped_kinds_exclude_administrative() -> None:
    assert log_kinds.is_search_ungrouped_kind("activation")
    assert log_kinds.is_search_ungrouped_kind("webdav")
    assert not log_kinds.is_search_ungrouped_kind("administrative")


def test_entry_render_kinds_include_administrative() -> None:
    assert log_kinds.is_entry_render_kind("administrative")
    assert log_kinds.is_entry_render_kind("activation")
    assert not log_kinds.is_entry_render_kind("smtp")


def test_entry_render_kinds_cover_search_ungrouped_kinds() -> None:
    missing = [
        kind
        for kind in log_kinds.SEARCH_UNGROUPED_KINDS
        if kind not in log_kinds.ENTRY_RENDER_KINDS
    ]
    assert not missing



"""Configuration loading for sm-logtool."""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
import os
from pathlib import Path
from typing import Any, Optional

import yaml

from .log_kinds import KIND_SMTP, normalize_kind


class ConfigError(Exception):
    """Raised when the configuration file cannot be parsed."""


_DEFAULT_CONFIG_ENV = "SM_LOGTOOL_CONFIG"
DEFAULT_LOGS_DIR = Path("/var/lib/smartermail/Logs")
DEFAULT_STAGING_DIR = Path("/var/tmp/sm-logtool/logs")
DEFAULT_KIND = KIND_SMTP
DEFAULT_THEME = "Cyberdark"
DEFAULT_THEME_MAPPING_PROFILE = "balanced"
DEFAULT_THEME_QUANTIZE_ANSI256 = True
SUPPORTED_THEME_MAPPING_PROFILES = (
    "balanced",
    "vivid",
    "soft",
)


@dataclass(frozen=True)
class AppConfig:
    """In-memory representation of sm-logtool configuration."""

    path: Path
    logs_dir: Optional[Path] = None
    staging_dir: Optional[Path] = None
    default_kind: str = KIND_SMTP
    theme: Optional[str] = None
    theme_import_paths: tuple[Path, ...] = ()
    theme_mapping_profile: str = DEFAULT_THEME_MAPPING_PROFILE
    theme_quantize_ansi256: bool = DEFAULT_THEME_QUANTIZE_ANSI256
    theme_overrides: dict[str, dict[str, str]] = field(default_factory=dict)

    @property
    def exists(self) -> bool:
        """Return ``True`` if the configuration file exists on disk."""

        return self.path.exists()


def default_config_path() -> Path:
    """Return the default config path, honoring ``SM_LOGTOOL_CONFIG``."""

    env_value = os.environ.get(_DEFAULT_CONFIG_ENV)
    if env_value:
        return Path(env_value).expanduser()
    return Path.home() / ".config" / "sm-logtool" / "config.yaml"


def load_config(path: Path | None = None) -> AppConfig:
    """Load configuration from ``path`` or the default location.

    When ``path`` is omitted, a default config file is created if missing.
    """

    config_path = (path or default_config_path()).expanduser()
    if path is None:
        _ensure_default_config_file(config_path)

    if not config_path.exists():
        return AppConfig(path=config_path)

    raw = _load_config_mapping(config_path)

    logs_dir = _coerce_path(raw.get("logs_dir"))
    staging_dir = _coerce_path(raw.get("staging_dir"))
    default_kind = raw.get("default_kind", DEFAULT_KIND)
    theme = raw.get("theme")
    theme_import_paths = _coerce_path_list(raw.get("theme_import_paths"))
    theme_mapping_profile = _coerce_theme_mapping_profile(
        raw.get("theme_mapping_profile", DEFAULT_THEME_MAPPING_PROFILE),
        config_path=config_path,
    )
    theme_quantize_ansi256 = _coerce_bool(
        raw.get(
            "theme_quantize_ansi256",
            DEFAULT_THEME_QUANTIZE_ANSI256,
        ),
        key="theme_quantize_ansi256",
        config_path=config_path,
    )
    theme_overrides = _coerce_theme_overrides(
        raw.get("theme_overrides", {}),
        config_path=config_path,
    )

    if not isinstance(default_kind, str):
        message = "Config key 'default_kind' must be a string"
        raise ConfigError(f"{message} (file: {config_path}).")
    default_kind = normalize_kind(default_kind)
    if theme is not None and not isinstance(theme, str):
        message = "Config key 'theme' must be a string"
        raise ConfigError(f"{message} (file: {config_path}).")

    return AppConfig(
        path=config_path,
        logs_dir=logs_dir,
        staging_dir=staging_dir,
        default_kind=default_kind,
        theme=theme,
        theme_import_paths=theme_import_paths,
        theme_mapping_profile=theme_mapping_profile,
        theme_quantize_ansi256=theme_quantize_ansi256,
        theme_overrides=theme_overrides,
    )


def save_theme(path: Path, theme: str) -> None:
    """Persist ``theme`` in a YAML config file at ``path``."""

    if not isinstance(theme, str) or not theme:
        raise ConfigError("Theme value must be a non-empty string.")

    config_path = path.expanduser()
    payload: dict[str, Any]
    if config_path.exists():
        payload = _load_config_mapping(config_path)
    else:
        payload = {}
    payload["theme"] = theme

    try:
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with config_path.open("w", encoding="utf-8") as handle:
            yaml.safe_dump(payload, handle, sort_keys=False)
    except OSError as exc:
        message = f"Failed to write config {config_path}: {exc}"
        raise ConfigError(message) from exc


def _ensure_default_config_file(config_path: Path) -> None:
    if config_path.exists():
        return
    payload = {
        "logs_dir": str(DEFAULT_LOGS_DIR),
        "staging_dir": str(DEFAULT_STAGING_DIR),
        "default_kind": DEFAULT_KIND,
        "theme": DEFAULT_THEME,
    }
    try:
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with config_path.open("w", encoding="utf-8") as handle:
            yaml.safe_dump(payload, handle, sort_keys=False)
    except OSError as exc:
        message = f"Failed to create default config {config_path}: {exc}"
        raise ConfigError(message) from exc


def _coerce_path(value: Any) -> Optional[Path]:
    if value is None:
        return None
    if isinstance(value, str):
        return Path(value).expanduser()
    typename = type(value).__name__
    raise ConfigError(
        f"Expected a string path in configuration, got {typename}."
    )


def _coerce_path_list(value: Any) -> tuple[Path, ...]:
    if value is None:
        return ()
    if isinstance(value, str):
        return (Path(value).expanduser(),)
    if isinstance(value, list):
        output: list[Path] = []
        for item in value:
            if not isinstance(item, str):
                typename = type(item).__name__
                raise ConfigError(
                    "Config key 'theme_import_paths' must contain string "
                    f"paths, got {typename}."
                )
            output.append(Path(item).expanduser())
        return tuple(output)
    typename = type(value).__name__
    raise ConfigError(
        "Config key 'theme_import_paths' must be a string path or list of "
        f"string paths, got {typename}."
    )


def _coerce_theme_mapping_profile(
    value: Any,
    *,
    config_path: Path,
) -> str:
    if not isinstance(value, str):
        typename = type(value).__name__
        raise ConfigError(
            "Config key 'theme_mapping_profile' must be a string, got "
            f"{typename} (file: {config_path})."
        )
    normalized = value.strip().lower()
    if normalized not in SUPPORTED_THEME_MAPPING_PROFILES:
        choices = ", ".join(SUPPORTED_THEME_MAPPING_PROFILES)
        raise ConfigError(
            "Config key 'theme_mapping_profile' must be one of "
            f"{choices} (file: {config_path})."
        )
    return normalized


def _coerce_bool(
    value: Any,
    *,
    key: str,
    config_path: Path,
) -> bool:
    if isinstance(value, bool):
        return value
    typename = type(value).__name__
    raise ConfigError(
        f"Config key '{key}' must be a boolean, got {typename} "
        f"(file: {config_path})."
    )


def _coerce_theme_overrides(
    value: Any,
    *,
    config_path: Path,
) -> dict[str, dict[str, str]]:
    if value is None:
        return {}
    if not isinstance(value, dict):
        typename = type(value).__name__
        raise ConfigError(
            "Config key 'theme_overrides' must be a mapping, got "
            f"{typename} (file: {config_path})."
        )

    output: dict[str, dict[str, str]] = {}
    for theme_name, mapping in value.items():
        if not isinstance(theme_name, str):
            typename = type(theme_name).__name__
            raise ConfigError(
                "Config key 'theme_overrides' must use string theme names, "
                f"got {typename} (file: {config_path})."
            )
        if not isinstance(mapping, dict):
            typename = type(mapping).__name__
            raise ConfigError(
                "Config key 'theme_overrides' values must be mappings, got "
                f"{typename} (file: {config_path})."
            )
        normalized: dict[str, str] = {}
        for color_key, raw in mapping.items():
            if not isinstance(color_key, str):
                typename = type(color_key).__name__
                raise ConfigError(
                    "Config key 'theme_overrides' color keys must be "
                    f"strings, got {typename} (file: {config_path})."
                )
            if not isinstance(raw, str):
                typename = type(raw).__name__
                raise ConfigError(
                    "Config key 'theme_overrides' color values must be "
                    f"strings, got {typename} (file: {config_path})."
                )
            normalized[color_key] = raw
        output[theme_name] = normalized
    return output


def _load_config_mapping(config_path: Path) -> dict[str, Any]:
    try:
        with config_path.open("r", encoding="utf-8") as handle:
            raw: Any = yaml.safe_load(handle) or {}
    except yaml.YAMLError as exc:  # pragma: no cover - parsing errors surfaced
        message = f"Failed to parse YAML config {config_path}: {exc}"
        raise ConfigError(message) from exc
    except OSError as exc:  # pragma: no cover - propagate filesystem errors
        message = f"Failed to read config {config_path}: {exc}"
        raise ConfigError(message) from exc

    if isinstance(raw, dict):
        return raw

    expected = type(raw).__name__
    message = (
        f"Expected a mapping at the top level of {config_path}, "
        f"got {expected}."
    )
    raise ConfigError(message)



"""Enforce the 79-character limit for production Python files."""

from __future__ import annotations

from pathlib import Path

MAX_LINE_LENGTH = 79
TARGET_DIRECTORIES = ("sm_logtool", "scripts")
LINE_LENGTH_EXCEPTIONS: dict[str, set[int]] = {}


def _line_length_violations(path: Path, root: Path) -> list[str]:
    relative_path = path.relative_to(root).as_posix()
    exceptions = LINE_LENGTH_EXCEPTIONS.get(relative_path, set())
    violations: list[str] = []

    lines = path.read_text(encoding="utf-8").splitlines()
    for line_number, line in enumerate(lines, 1):
        if len(line) <= MAX_LINE_LENGTH:
            continue
        if line_number in exceptions:
            continue
        length = len(line)
        violations.append(
            f"{relative_path}:{line_number} has {length} characters",
        )
    return violations


def test_production_python_line_lengths() -> None:
    project_root = Path(__file__).resolve().parents[1]
    violations: list[str] = []

    for directory in TARGET_DIRECTORIES:
        base = project_root / directory
        for path in sorted(base.rglob("*.py")):
            violations.extend(_line_length_violations(path, project_root))

    assert not violations, (
        "Production line-length violations detected:\n"
        + "\n".join(violations)
    )



from sm_logtool.ui.app import _osc52_sequence
from sm_logtool.ui.app import LogBrowser


def test_osc52_sequence_uses_standard_st_terminator():
    sequence = _osc52_sequence("YWJj", env={})
    assert sequence.startswith("\x1b]52;c;YWJj")
    assert sequence.endswith("\x1b\\")


def test_osc52_sequence_wraps_for_tmux():
    sequence = _osc52_sequence("YWJj", env={"TMUX": "/tmp/tmux-1000/default"})
    assert sequence.startswith("\x1bPtmux;\x1b\x1b]52;c;YWJj")
    assert sequence.endswith("\x1b\\\x1b\\")


def test_osc52_sequence_wraps_for_screen():
    sequence = _osc52_sequence("YWJj", env={"TERM": "screen-256color"})
    assert sequence.startswith("\x1bP\x1b]52;c;YWJj")
    assert sequence.endswith("\x1b\\\x1b\\")


def test_copy_status_message_for_large_payload(tmp_path):
    logs_dir = tmp_path / "logs"
    logs_dir.mkdir()
    app = LogBrowser(logs_dir=logs_dir)
    message = app._copy_status_message(
        selection_only=True,
        mode="terminal-large",
    )
    assert "Payload is large" in message



"""Command-line entry point for sm-logtool.

Provides a TUI browser (`browse`) and a search workflow (`search`).
"""

from __future__ import annotations

import argparse
from collections.abc import Iterable
from dataclasses import dataclass
from importlib import metadata
import os
import textwrap
from pathlib import Path
import sys
from typing import Callable

from rich.console import Console

from .config import AppConfig, ConfigError, load_config
from .highlighting import highlight_line
from .log_kinds import SUPPORTED_KINDS, normalize_kind
from .logfiles import (
    UnknownLogDate,
    find_log_by_date,
    newest_log,
    parse_log_filename,
    parse_stamp,
    summarize_logs,
)
from .result_rendering import render_search_results
from .result_modes import normalize_result_mode
from .result_modes import RESULT_MODE_DESCRIPTIONS
from .result_modes import RESULT_MODE_RELATED_TRAFFIC
from .result_modes import SUPPORTED_RESULT_MODES
from .search_modes import (
    DEFAULT_FUZZY_THRESHOLD,
    MODE_LITERAL,
    SEARCH_MODE_DESCRIPTIONS,
    SUPPORTED_SEARCH_MODES,
    normalize_fuzzy_threshold,
    normalize_search_mode,
)
from .search import get_search_function
from .staging import stage_log
from .ui.theme_importer import ensure_default_theme_dirs
from .ui.theme_importer import SUPPORTED_THEME_MAPPING_PROFILES


CONFIG_ATTR = "_config"


@dataclass(frozen=True)
class _SearchOptions:
    mode: str
    fuzzy_threshold: float
    result_mode: str


@dataclass(frozen=True)
class _SearchContext:
    logs_dir: Path
    staging_dir: Path
    log_kind: str


@dataclass(frozen=True)
class _SearchRunResult:
    results: list
    targets: list[Path]
    log_kind: str
    result_mode: str


def build_parser() -> argparse.ArgumentParser:
    """Create the top-level argument parser."""

    parser = _build_root_parser()
    subparsers = parser.add_subparsers(dest="command")
    _add_browse_subcommand(subparsers)
    _add_themes_subcommand(subparsers)
    _add_search_subcommand(subparsers)
    parser.set_defaults(command="browse", handler=_run_browse)
    return parser


def _build_root_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="sm-logtool",
        description=_parser_description(),
    )
    parser.add_argument(
        "--config",
        type=Path,
        default=None,
        help=(
            "Path to a YAML config file. Defaults to $SM_LOGTOOL_CONFIG or "
            "~/.config/sm-logtool/config.yaml."
        ),
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {_package_version()}",
    )
    return parser


def _parser_description() -> str:
    return textwrap.dedent(
        """
        Explore SmarterMail logs from your terminal.
        The `browse` subcommand launches the Textual UI.
        The `search` subcommand performs a console-based search across
        supported SmarterMail log kinds.
        The `themes` subcommand launches the visual theme converter.

        Config-aware defaults:
          - logs_dir and staging_dir come from config.yaml when present.
          - Command-line flags override config values.
        """
    ).strip()


def _add_browse_subcommand(subparsers: argparse._SubParsersAction) -> None:
    browse_parser = subparsers.add_parser(
        "browse",
        help="Launch the Textual UI",
    )
    browse_parser.add_argument(
        "--logs-dir",
        type=Path,
        default=None,
        help=(
            "Path containing SmarterMail log files. Optional when logs_dir "
            "is set in config.yaml."
        ),
    )
    browse_parser.set_defaults(handler=_run_browse)


def _add_themes_subcommand(subparsers: argparse._SubParsersAction) -> None:
    themes_parser = subparsers.add_parser(
        "themes",
        help="Open visual theme conversion utility",
    )
    themes_parser.add_argument(
        "--source",
        type=Path,
        action="append",
        default=None,
        help=(
            "Theme source file/directory. Repeatable. Supports "
            ".itermcolors/.colors/.colortheme."
        ),
    )
    themes_parser.add_argument(
        "--store-dir",
        type=Path,
        default=None,
        help=(
            "Directory for converted sm-logtool themes. Defaults to "
            "~/.config/sm-logtool/themes."
        ),
    )
    themes_parser.add_argument(
        "--profile",
        choices=SUPPORTED_THEME_MAPPING_PROFILES,
        default="balanced",
        help="Default mapping profile for preview/save.",
    )
    themes_parser.add_argument(
        "--no-ansi256",
        action="store_true",
        help="Disable ANSI-256 quantization and keep truecolor output.",
    )
    themes_parser.set_defaults(handler=_run_themes)


def _add_search_subcommand(subparsers: argparse._SubParsersAction) -> None:
    search_parser = _new_search_parser(subparsers)
    _add_search_term_argument(search_parser)
    _add_search_mode_arguments(search_parser)
    _add_search_target_arguments(search_parser)
    _add_search_control_arguments(search_parser)
    search_parser.set_defaults(handler=_run_search)


def _new_search_parser(
    subparsers: argparse._SubParsersAction,
) -> argparse.ArgumentParser:
    return subparsers.add_parser(
        "search",
        help="Search SmarterMail logs for a term",
        description=(
            "Search a SmarterMail log kind using the selected search mode."
        ),
        epilog=_search_help_epilog(),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )


def _add_search_term_argument(search_parser: argparse.ArgumentParser) -> None:
    search_parser.add_argument(
        "term",
        nargs="?",
        default=None,
        help=(
            "Search pattern (case-insensitive by default). Pattern syntax "
            "depends on --mode."
        ),
    )


def _add_search_mode_arguments(search_parser: argparse.ArgumentParser) -> None:
    search_parser.add_argument(
        "--mode",
        choices=SUPPORTED_SEARCH_MODES,
        default=MODE_LITERAL,
        help=(
            "Search mode to use. "
            "literal=substring, wildcard supports '*' and '?', "
            "regex=Python re syntax (PCRE-like, not full PCRE), "
            "fuzzy=approximate similarity matching."
        ),
    )
    search_parser.add_argument(
        "--fuzzy-threshold",
        type=float,
        default=DEFAULT_FUZZY_THRESHOLD,
        help=(
            "Similarity threshold for --mode fuzzy "
            f"({0.0:.2f} to {1.0:.2f}, default {DEFAULT_FUZZY_THRESHOLD:.2f})."
        ),
    )
    search_parser.add_argument(
        "--result-mode",
        choices=SUPPORTED_RESULT_MODES,
        default=RESULT_MODE_RELATED_TRAFFIC,
        help=(
            "Result rendering mode. "
            "related=show full grouped conversations, "
            "matching-only=show only directly matching lines."
        ),
    )


def _add_search_target_arguments(
    search_parser: argparse.ArgumentParser,
) -> None:
    search_parser.add_argument(
        "--logs-dir",
        type=Path,
        default=None,
        help=(
            "Directory containing source logs. Optional when logs_dir is "
            "set in config.yaml."
        ),
    )
    search_parser.add_argument(
        "--staging-dir",
        type=Path,
        default=None,
        help=(
            "Directory where logs are copied/extracted before analysis. "
            "Optional when staging_dir is set in config.yaml."
        ),
    )
    search_parser.add_argument(
        "--log-file",
        type=Path,
        action="append",
        default=None,
        help=(
            "Specific log file to search. Relative paths resolve under "
            "--logs-dir. Repeat to search multiple files."
        ),
    )
    search_parser.add_argument(
        "--kind",
        default=None,
        help=(
            "Log kind to search. Optional when default_kind is set in "
            "config.yaml. "
            "Use --list-kinds to show supported values."
        ),
    )
    search_parser.add_argument(
        "--date",
        action="append",
        default=None,
        help=(
            "Date (YYYY.MM.DD) of the log file to search. Repeat to search "
            "multiple dates."
        ),
    )


def _add_search_control_arguments(
    search_parser: argparse.ArgumentParser,
) -> None:
    search_parser.add_argument(
        "--list",
        action="store_true",
        help="List available logs for the selected kind and exit.",
    )
    search_parser.add_argument(
        "--list-kinds",
        action="store_true",
        help="List supported log kinds and exit.",
    )
    search_parser.add_argument(
        "--case-sensitive",
        action="store_true",
        help="Treat the search term as case-sensitive.",
    )


def main(argv: list[str] | None = None) -> int:
    """Entry point used by console scripts and ``python -m sm_logtool.cli``."""

    parser = build_parser()
    args = parser.parse_args(argv)

    try:
        config = load_config(args.config)
    except ConfigError as exc:
        parser.error(str(exc))

    setattr(args, CONFIG_ATTR, config)

    handler: Callable[[argparse.Namespace], int]
    handler = getattr(args, "handler", _run_browse)

    return handler(args)


def _package_version() -> str:
    try:
        return metadata.version("sm-logtool")
    except metadata.PackageNotFoundError:
        return "unknown"


def _run_browse(args: argparse.Namespace) -> int:
    config: AppConfig = getattr(args, CONFIG_ATTR)
    try:
        logs_dir = _resolve_logs_dir(args, config)
        staging_dir = _resolve_staging_dir(args, config)
    except ValueError as exc:
        print(str(exc), file=sys.stderr)
        return 2

    # Lazy import so tests can import this module without textual installed.
    try:
        from .ui.app import run as run_tui  # type: ignore
    except Exception as exc:  # pragma: no cover - defensive fallback
        raise SystemExit(
            "The Textual UI could not be loaded. Ensure the 'textual' package "
            "is installed.\n"
            f"Details: {exc}"
        ) from exc
    theme_store_dir, _ = ensure_default_theme_dirs(config.path)

    return run_tui(
        logs_dir,
        staging_dir=staging_dir,
        default_kind=config.default_kind,
        config_path=config.path,
        theme=config.theme,
        theme_store_dir=theme_store_dir,
        theme_import_paths=config.theme_import_paths,
        theme_mapping_profile=config.theme_mapping_profile,
        theme_quantize_ansi256=config.theme_quantize_ansi256,
        theme_overrides=config.theme_overrides,
        persist_theme_changes=_should_persist_theme_changes(args),
    )


def _run_themes(args: argparse.Namespace) -> int:
    config: AppConfig = getattr(args, CONFIG_ATTR)
    default_store_dir, default_source_dir = ensure_default_theme_dirs(
        config.path
    )
    source_paths = tuple(
        args.source
        or (default_source_dir,)
    )
    store_dir = args.store_dir or default_store_dir

    try:
        from .ui.theme_studio import run as run_studio  # type: ignore
    except Exception as exc:  # pragma: no cover - defensive fallback
        raise SystemExit(
            "The theme studio UI could not be loaded. Ensure the 'textual' "
            "package is installed.\n"
            f"Details: {exc}"
        ) from exc

    return run_studio(
        source_paths=source_paths,
        store_dir=store_dir,
        profile=args.profile,
        quantize_ansi256=not args.no_ansi256,
    )


def _should_persist_theme_changes(args: argparse.Namespace) -> bool:
    if args.config is not None:
        return False
    if os.environ.get("SM_LOGTOOL_CONFIG"):
        return False
    return True


def _run_search(args: argparse.Namespace) -> int:
    config: AppConfig = getattr(args, CONFIG_ATTR)
    if getattr(args, "list_kinds", False):
        return _list_kinds()

    try:
        outcome = _run_search_flow(args, config)
    except ValueError as exc:
        print(str(exc), file=sys.stderr)
        return 2
    except RuntimeError as exc:
        print(str(exc), file=sys.stderr)
        return 1
    if isinstance(outcome, int):
        return outcome

    _print_search_summary(
        outcome.results,
        outcome.targets,
        outcome.log_kind,
        result_mode=outcome.result_mode,
    )
    return 0


def _run_search_flow(
    args: argparse.Namespace,
    config: AppConfig,
) -> int | _SearchRunResult:
    options = _resolve_search_options(args)
    context = _resolve_search_context(args, config)
    if args.list:
        return _list_logs(context.logs_dir, context.log_kind)
    term = _resolve_search_term(args)
    search_fn = _resolve_search_function(context.log_kind)
    targets = _resolve_search_targets_or_raise(
        args,
        context.logs_dir,
        context.log_kind,
    )
    _ensure_targets_available(
        targets,
        log_kind=context.log_kind,
        logs_dir=context.logs_dir,
    )
    results = _execute_search_targets(
        search_fn,
        targets,
        term=term,
        staging_dir=context.staging_dir,
        mode=options.mode,
        fuzzy_threshold=options.fuzzy_threshold,
        ignore_case=not args.case_sensitive,
    )
    return _SearchRunResult(
        results=results,
        targets=targets,
        log_kind=context.log_kind,
        result_mode=options.result_mode,
    )


def _resolve_search_options(args: argparse.Namespace) -> _SearchOptions:
    return _SearchOptions(
        mode=normalize_search_mode(getattr(args, "mode", MODE_LITERAL)),
        fuzzy_threshold=normalize_fuzzy_threshold(
            getattr(args, "fuzzy_threshold", DEFAULT_FUZZY_THRESHOLD),
        ),
        result_mode=normalize_result_mode(
            getattr(args, "result_mode", RESULT_MODE_RELATED_TRAFFIC),
        ),
    )


def _resolve_search_context(
    args: argparse.Namespace,
    config: AppConfig,
) -> _SearchContext:
    logs_dir = _resolve_logs_dir(args, config)
    staging_dir = _resolve_staging_dir(args, config)
    log_kind_value = args.kind or config.default_kind
    if log_kind_value is None:
        raise ValueError("Log kind is required.")
    return _SearchContext(
        logs_dir=logs_dir,
        staging_dir=staging_dir,
        log_kind=normalize_kind(log_kind_value),
    )


def _resolve_search_term(args: argparse.Namespace) -> str:
    term = args.term
    if term is None:
        raise ValueError("Search term is required unless --list is supplied.")
    return term


def _resolve_search_function(log_kind: str):
    search_fn = get_search_function(log_kind)
    if search_fn is None:
        raise ValueError(f"Unsupported log kind: {log_kind}")
    return search_fn


def _resolve_search_targets_or_raise(
    args: argparse.Namespace,
    logs_dir: Path,
    log_kind: str,
) -> list[Path]:
    try:
        return _resolve_search_targets(args, logs_dir, log_kind)
    except (UnknownLogDate, ValueError) as exc:
        raise ValueError(str(exc)) from exc


def _ensure_targets_available(
    targets: list[Path],
    *,
    log_kind: str,
    logs_dir: Path,
) -> None:
    if targets:
        return
    raise ValueError(f"No {log_kind} logs found in {logs_dir}")


def _execute_search_targets(
    search_fn,
    targets: list[Path],
    *,
    term: str,
    staging_dir: Path,
    mode: str,
    fuzzy_threshold: float,
    ignore_case: bool,
) -> list:
    results = []
    for source_path in targets:
        staged_path = _stage_search_target(
            source_path,
            staging_dir=staging_dir,
        )
        result = search_fn(
            staged_path,
            term,
            mode=mode,
            fuzzy_threshold=fuzzy_threshold,
            ignore_case=ignore_case,
        )
        results.append(result)
    return results


def _stage_search_target(
    source_path: Path,
    *,
    staging_dir: Path,
) -> Path:
    try:
        staged = stage_log(
            source_path,
            staging_dir=staging_dir,
        )
    except Exception as exc:  # pragma: no cover - surface staging failure
        message = f"Failed to stage log {source_path}: {exc}"
        raise RuntimeError(message) from exc
    return staged.staged_path


def _print_search_summary(
    results,
    source_paths: list[Path],
    log_kind: str,
    *,
    result_mode: str = RESULT_MODE_RELATED_TRAFFIC,
) -> None:
    console = _build_stdout_console()
    lines = render_search_results(
        results,
        source_paths,
        log_kind,
        result_mode=result_mode,
    )
    for line in lines:
        _write_highlighted(console, log_kind, line)


def _build_stdout_console() -> Console:
    return Console(highlight=False, soft_wrap=True)


def _write_highlighted(
    console: Console,
    log_kind: str,
    line: str,
) -> None:
    if not line:
        console.print()
        return
    console.print(highlight_line(log_kind, line))


def _list_logs(logs_dir: Path, kind: str) -> int:
    resolved_kind = normalize_kind(kind)
    logs = summarize_logs(logs_dir, resolved_kind)
    if not logs:
        print(f"No {resolved_kind} logs found in {logs_dir}", file=sys.stderr)
        return 2

    print(f"Available {resolved_kind} logs in {logs_dir}:")
    for info in logs:
        stamp = info.stamp.strftime('%Y.%m.%d') if info.stamp else 'unknown'
        suffix = ' (zip)' if info.is_zipped else ''
        print(f"  {stamp} -> {info.path.name}{suffix}")
    return 0


def _list_kinds() -> int:
    print("Supported log kinds:")
    for kind in SUPPORTED_KINDS:
        print(f"  {kind}")
    return 0


def _resolve_logs_dir(args: argparse.Namespace, config: AppConfig) -> Path:
    candidate = getattr(args, "logs_dir", None) or config.logs_dir
    if candidate is None:
        raise ValueError(
            "Log directory is not configured. Set 'logs_dir' in config.yaml "
            "or pass --logs-dir."
        )
    return candidate


def _resolve_staging_dir(
    args: argparse.Namespace,
    config: AppConfig,
) -> Path:
    candidate = getattr(args, "staging_dir", None) or config.staging_dir
    if candidate is None:
        raise ValueError(
            "Staging directory is not configured. Set 'staging_dir' in "
            "config.yaml or pass --staging-dir."
        )
    return candidate


def _resolve_search_targets(
    args: argparse.Namespace,
    logs_dir: Path,
    log_kind: str,
) -> list[Path]:
    log_files = _normalize_path_values(getattr(args, "log_file", None))
    date_values = _normalize_text_values(getattr(args, "date", None))
    if log_files and date_values:
        raise ValueError("--log-file and --date cannot be used together.")

    if log_files:
        return _resolve_log_file_targets(logs_dir, log_kind, log_files)
    if date_values:
        return _resolve_date_targets(logs_dir, log_kind, date_values)
    newest = newest_log(logs_dir, log_kind)
    return [newest.path] if newest is not None else []


def _resolve_log_file_targets(
    logs_dir: Path,
    log_kind: str,
    log_files: list[Path],
) -> list[Path]:
    targets: list[Path] = []
    seen: set[Path] = set()
    kind_key = normalize_kind(log_kind)
    for value in log_files:
        target = value if value.is_absolute() else logs_dir / value
        if not target.exists():
            raise ValueError(f"Log file not found: {target}")
        parsed = parse_log_filename(target)
        if parsed.kind and parsed.kind != kind_key:
            raise ValueError(
                f"Log file {target.name} does not match kind {kind_key}."
            )
        resolved = target.resolve()
        if resolved in seen:
            continue
        seen.add(resolved)
        targets.append(target)
    return targets


def _resolve_date_targets(
    logs_dir: Path,
    log_kind: str,
    date_values: list[str],
) -> list[Path]:
    targets: list[Path] = []
    seen: set[Path] = set()
    for value in date_values:
        target_date = parse_stamp(value)
        info = find_log_by_date(logs_dir, log_kind, target_date)
        if info is None:
            raise ValueError(
                f"No {log_kind} log found for {target_date:%Y.%m.%d} in "
                f"{logs_dir}"
            )
        resolved = info.path.resolve()
        if resolved in seen:
            continue
        seen.add(resolved)
        targets.append(info.path)
    return targets


def _normalize_path_values(values: object) -> list[Path]:
    if values is None:
        return []
    if isinstance(values, Path):
        return [values]
    if not isinstance(values, Iterable):
        raise TypeError("Expected iterable path values.")
    output: list[Path] = []
    for value in values:
        if not isinstance(value, Path):
            raise TypeError("Expected Path values.")
        output.append(value)
    return output


def _normalize_text_values(values: object) -> list[str]:
    if values is None:
        return []
    if isinstance(values, str):
        return [values]
    if not isinstance(values, Iterable):
        raise TypeError("Expected iterable text values.")
    output: list[str] = []
    for value in values:
        if not isinstance(value, str):
            raise TypeError("Expected string values.")
        output.append(value)
    return output


def _search_help_epilog() -> str:
    kinds = ", ".join(SUPPORTED_KINDS)
    mode_lines = "\n".join(
        f"  - {mode}: {SEARCH_MODE_DESCRIPTIONS[mode]}"
        for mode in SUPPORTED_SEARCH_MODES
    )
    result_mode_lines = "\n".join(
        f"  - {mode}: {RESULT_MODE_DESCRIPTIONS[mode]}"
        for mode in SUPPORTED_RESULT_MODES
    )
    return textwrap.dedent(
        f"""
        Target resolution:
          1. If --log-file is provided (repeatable), those files are searched.
          2. Else if --date is provided (repeatable), those dates are searched.
          3. Else the newest available log for --kind is searched.

        Search modes:
{mode_lines}

        Result modes:
{result_mode_lines}

        Available kinds:
          {kinds}
        """
    ).strip()


def scan_logs(logs_dir: Path) -> list[Path]:
    """Return log files underneath ``logs_dir`` sorted by name."""

    if not logs_dir.exists():
        raise FileNotFoundError(f"Logs directory not found: {logs_dir}")

    if not logs_dir.is_dir():
        raise NotADirectoryError(f"Logs path is not a directory: {logs_dir}")

    return sorted(
        path
        for path in logs_dir.iterdir()
        if path.is_file() and not path.name.startswith('.')
    )


if __name__ == "__main__":
    raise SystemExit(main())



from sm_logtool.search_planning import choose_search_execution_plan


def test_plan_uses_serial_for_single_target():
    plan = choose_search_execution_plan(
        1,
        500_000_000,
        use_index_cache=False,
        max_workers=4,
    )
    assert plan.workers == 1
    assert "single target" in plan.reason


def test_plan_uses_serial_for_small_two_target_indexed_workload():
    plan = choose_search_execution_plan(
        2,
        500_000_000,
        use_index_cache=True,
        max_workers=4,
    )
    assert plan.workers == 1
    assert "indexed two-target" in plan.reason


def test_plan_uses_parallel_for_large_indexed_workload():
    plan = choose_search_execution_plan(
        6,
        2 * 1024 * 1024 * 1024,
        use_index_cache=True,
        max_workers=4,
    )
    assert plan.workers == 4
    assert "indexed large workload" in plan.reason


def test_plan_uses_serial_for_small_two_target_workload():
    plan = choose_search_execution_plan(
        2,
        64 * 1024 * 1024,
        use_index_cache=False,
        max_workers=4,
    )
    assert plan.workers == 1
    assert "small two-target" in plan.reason


def test_plan_uses_serial_for_small_per_target_workload():
    plan = choose_search_execution_plan(
        3,
        100 * 1024 * 1024,
        use_index_cache=False,
        max_workers=4,
    )
    assert plan.workers == 1
    assert "small per-target" in plan.reason


def test_plan_limits_medium_workload_parallelism():
    plan = choose_search_execution_plan(
        6,
        300 * 1024 * 1024,
        use_index_cache=False,
        max_workers=4,
    )
    assert plan.workers == 2
    assert "medium workload" in plan.reason


def test_plan_uses_bounded_workers_for_large_workload():
    plan = choose_search_execution_plan(
        8,
        2 * 1024 * 1024 * 1024,
        use_index_cache=False,
        max_workers=4,
    )
    assert plan.workers == 4
    assert "large workload" in plan.reason



from __future__ import annotations

from datetime import date

from sm_logtool import logfiles


def test_find_log_by_date_returns_matching_file(tmp_path):
    logs_dir = tmp_path / "logs"
    logs_dir.mkdir()
    (logs_dir / "2024.01.01-smtpLog.log").write_text("\n")
    (logs_dir / "2024.01.02-smtpLog.log").write_text("\n")

    info = logfiles.find_log_by_date(logs_dir, "smtp", date(2024, 1, 2))
    assert info is not None
    assert info.path.name == "2024.01.02-smtpLog.log"


def test_discover_logs_sorts_newest_first(tmp_path):
    logs_dir = tmp_path / "logs"
    logs_dir.mkdir()
    (logs_dir / "2024.01.01-smtpLog.log").write_text("\n")
    (logs_dir / "2024.01.03-smtpLog.log.zip").write_text("fake zip")
    (logs_dir / "2024.01.02-smtpLog.log").write_text("\n")

    infos = logfiles.discover_logs(logs_dir, "smtp")
    assert [info.path.name for info in infos] == [
        "2024.01.03-smtpLog.log.zip",
        "2024.01.02-smtpLog.log",
        "2024.01.01-smtpLog.log",
    ]


def test_parse_stamp_rejects_invalid_value():
    try:
        logfiles.parse_stamp("2024-01-01")
    except logfiles.UnknownLogDate as exc:
        assert "Invalid log date stamp" in str(exc)
    else:
        raise AssertionError("Expected UnknownLogDate to be raised")



"""Import terminal color themes as Textual themes."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
import plistlib
import re
from typing import Iterable, Mapping

from rich.color import Color
from rich.color_triplet import ColorTriplet
from textual.theme import Theme
import yaml

from .themes import CYBER_THEME_VARIABLE_DEFAULTS

SUPPORTED_THEME_IMPORT_SUFFIXES = (
    ".itermcolors",
    ".colors",
    ".colortheme",
)
SUPPORTED_THEME_MAPPING_PROFILES = (
    "balanced",
    "vivid",
    "soft",
)
THEME_FILE_SUFFIX = ".smlogtheme.yaml"

_SEMANTIC_COLOR_KEYS = {
    "primary",
    "secondary",
    "warning",
    "error",
    "success",
    "accent",
    "foreground",
    "background",
    "surface",
    "panel",
}

_DEFAULT_ANSI = (
    ColorTriplet(0, 0, 0),
    ColorTriplet(205, 0, 0),
    ColorTriplet(0, 205, 0),
    ColorTriplet(205, 205, 0),
    ColorTriplet(0, 0, 238),
    ColorTriplet(205, 0, 205),
    ColorTriplet(0, 205, 205),
    ColorTriplet(229, 229, 229),
    ColorTriplet(127, 127, 127),
    ColorTriplet(255, 0, 0),
    ColorTriplet(0, 255, 0),
    ColorTriplet(255, 255, 0),
    ColorTriplet(92, 92, 255),
    ColorTriplet(255, 0, 255),
    ColorTriplet(0, 255, 255),
    ColorTriplet(255, 255, 255),
)


@dataclass(frozen=True)
class TerminalPalette:
    """Terminal palette values used for semantic theme mapping."""

    name: str
    source: Path
    background: ColorTriplet
    foreground: ColorTriplet
    cursor: ColorTriplet
    ansi: tuple[ColorTriplet, ...]


@dataclass(frozen=True)
class _ProfileSpec:
    primary_slots: tuple[int, ...]
    secondary_slots: tuple[int, ...]
    success_slots: tuple[int, ...]
    warning_slots: tuple[int, ...]
    error_slots: tuple[int, ...]
    accent_slots: tuple[int, ...]
    panel_mix: float
    surface_mix: float
    action_mix: float
    selection_mix: float
    active_mix: float
    button_mix: float


_PROFILE_SPECS = {
    "balanced": _ProfileSpec(
        primary_slots=(14, 12, 6, 4),
        secondary_slots=(13, 5, 12, 4),
        success_slots=(10, 2),
        warning_slots=(11, 3),
        error_slots=(9, 1),
        accent_slots=(13, 11, 10, 9, 15),
        panel_mix=0.08,
        surface_mix=0.04,
        action_mix=0.14,
        selection_mix=0.30,
        active_mix=0.35,
        button_mix=0.20,
    ),
    "vivid": _ProfileSpec(
        primary_slots=(12, 14, 13, 9),
        secondary_slots=(14, 12, 11, 10),
        success_slots=(10, 14, 2),
        warning_slots=(11, 13, 3),
        error_slots=(9, 13, 1),
        accent_slots=(13, 11, 10, 9, 14),
        panel_mix=0.16,
        surface_mix=0.08,
        action_mix=0.32,
        selection_mix=0.48,
        active_mix=0.58,
        button_mix=0.34,
    ),
    "soft": _ProfileSpec(
        primary_slots=(6, 4, 14, 12),
        secondary_slots=(7, 8, 6, 4),
        success_slots=(2, 10, 6),
        warning_slots=(3, 11, 5),
        error_slots=(1, 9, 5),
        accent_slots=(5, 6, 4, 3),
        panel_mix=0.05,
        surface_mix=0.02,
        action_mix=0.08,
        selection_mix=0.18,
        active_mix=0.22,
        button_mix=0.12,
    ),
}

def load_imported_themes(
    paths: Iterable[Path],
    *,
    profile: str,
    overrides: Mapping[str, Mapping[str, str]] | None = None,
    quantize_ansi256: bool,
    existing_names: set[str] | None = None,
) -> tuple[list[Theme], list[str]]:
    """Load themes from files/directories and map them to Textual themes."""

    normalized_profile = normalize_mapping_profile(profile)
    override_map = _normalize_override_map(overrides)
    registered_names = set(existing_names or ())
    files = discover_theme_files(paths)

    themes: list[Theme] = []
    warnings: list[str] = []
    for file_path in files:
        try:
            palette = parse_terminal_palette(file_path)
        except ValueError as exc:
            warnings.append(f"{file_path}: {exc}")
            continue

        source_name = palette.name
        theme_name = _unique_theme_name(source_name, registered_names)
        resolved = (
            override_map.get(theme_name)
            or override_map.get(source_name)
        )
        mapped = map_terminal_palette(
            name=theme_name,
            palette=palette,
            profile=normalized_profile,
            overrides=resolved,
            quantize_ansi256=quantize_ansi256,
        )
        themes.append(mapped)
    return themes, warnings


def discover_theme_files(paths: Iterable[Path]) -> list[Path]:
    """Return sorted importable theme files from configured paths."""

    files: set[Path] = set()
    suffixes = set(SUPPORTED_THEME_IMPORT_SUFFIXES)
    for raw_path in paths:
        path = raw_path.expanduser()
        if not path.exists():
            continue
        if path.is_file() and path.suffix.lower() in suffixes:
            files.add(path)
            continue
        if not path.is_dir():
            continue
        for child in path.rglob("*"):
            if not child.is_file():
                continue
            if child.suffix.lower() in suffixes:
                files.add(child)
    return sorted(files)


def normalize_mapping_profile(profile: str) -> str:
    """Normalize and validate a mapping profile name."""

    normalized = profile.strip().lower()
    if normalized not in SUPPORTED_THEME_MAPPING_PROFILES:
        choices = ", ".join(SUPPORTED_THEME_MAPPING_PROFILES)
        raise ValueError(
            f"Unsupported theme_mapping_profile '{profile}'. "
            f"Expected one of: {choices}."
        )
    return normalized


def parse_terminal_palette(path: Path) -> TerminalPalette:
    """Parse a terminal palette from ``path``."""

    suffix = path.suffix.lower()
    if suffix == ".itermcolors":
        return _parse_itermcolors(path)
    return _parse_line_theme(path)


def map_terminal_palette(
    *,
    name: str,
    palette: TerminalPalette,
    profile: str,
    overrides: Mapping[str, str] | None,
    quantize_ansi256: bool,
) -> Theme:
    """Convert a terminal palette into a Textual app theme."""

    profile_name = normalize_mapping_profile(profile)
    spec = _PROFILE_SPECS[profile_name]
    dark = _luminance(palette.background) < 0.45

    background = palette.background
    foreground = _ensure_contrast(
        palette.foreground,
        background,
        minimum_ratio=7.0,
        prefer_light=dark,
    )

    primary = _pick_slot_color(
        palette,
        spec.primary_slots,
        background,
        minimum_ratio=3.0,
        prefer_light=dark,
    )
    secondary = _pick_slot_color(
        palette,
        spec.secondary_slots,
        background,
        minimum_ratio=3.0,
        prefer_light=dark,
    )
    warning = _pick_slot_color(
        palette,
        spec.warning_slots,
        background,
        minimum_ratio=3.0,
        prefer_light=dark,
    )
    error = _pick_slot_color(
        palette,
        spec.error_slots,
        background,
        minimum_ratio=3.0,
        prefer_light=dark,
    )
    success = _pick_slot_color(
        palette,
        spec.success_slots,
        background,
        minimum_ratio=3.0,
        prefer_light=dark,
    )
    accent = _pick_slot_color(
        palette,
        spec.accent_slots,
        background,
        minimum_ratio=3.0,
        prefer_light=dark,
    )

    surface = _ensure_contrast(
        _blend(background, foreground, spec.surface_mix),
        foreground,
        minimum_ratio=4.5,
        prefer_light=not dark,
    )
    panel = _ensure_contrast(
        _blend(background, foreground, spec.panel_mix),
        foreground,
        minimum_ratio=4.5,
        prefer_light=not dark,
    )

    semantic: dict[str, ColorTriplet] = {
        "primary": primary,
        "secondary": secondary,
        "warning": warning,
        "error": error,
        "success": success,
        "accent": accent,
        "foreground": foreground,
        "background": background,
        "surface": surface,
        "panel": panel,
    }

    variables = _derive_theme_variables(
        semantic,
        action_mix=spec.action_mix,
        selection_mix=spec.selection_mix,
        active_mix=spec.active_mix,
        button_mix=spec.button_mix,
        dark=dark,
    )
    _apply_overrides(semantic, variables, overrides or {}, palette)

    if quantize_ansi256:
        semantic = {
            key: _nearest_xterm_256(color)
            for key, color in semantic.items()
        }
        variables = {
            key: _as_hex(
                _nearest_xterm_256(
                    Color.parse(value).get_truecolor()
                )
                )
            for key, value in variables.items()
        }

    _ensure_distinct_selection_states(
        variables=variables,
        semantic=semantic,
        dark=dark,
        quantize_ansi256=quantize_ansi256,
    )

    return Theme(
        name=name,
        primary=_as_hex(semantic["primary"]),
        secondary=_as_hex(semantic["secondary"]),
        warning=_as_hex(semantic["warning"]),
        error=_as_hex(semantic["error"]),
        success=_as_hex(semantic["success"]),
        accent=_as_hex(semantic["accent"]),
        foreground=_as_hex(semantic["foreground"]),
        background=_as_hex(semantic["background"]),
        surface=_as_hex(semantic["surface"]),
        panel=_as_hex(semantic["panel"]),
        dark=dark,
        variables=variables,
    )


def _ensure_distinct_selection_states(
    *,
    variables: dict[str, str],
    semantic: Mapping[str, ColorTriplet],
    dark: bool,
    quantize_ansi256: bool,
) -> None:
    keys = (
        "selection-selected-background",
        "selection-active-background",
        "selection-selected-active-background",
    )
    defaults = (
        _blend(semantic["background"], semantic["accent"], 0.30),
        _blend(semantic["background"], semantic["primary"], 0.35),
        _blend(semantic["background"], semantic["warning"], 0.35),
    )
    values = [
        _parse_variable_color(variables.get(key), fallback)
        for key, fallback in zip(keys, defaults)
    ]
    if _selection_states_are_distinct(values):
        return

    candidates = _selection_candidates(
        semantic=semantic,
        dark=dark,
        quantize_ansi256=quantize_ansi256,
    )
    resolved: list[ColorTriplet] = []
    for index, current in enumerate(values):
        preferred = (
            current
            if _is_color_distinct_from(
                current,
                resolved,
                minimum_distance=900,
            )
            else None
        )
        if preferred is None:
            preferred = _pick_distinct_candidate(
                candidates=candidates,
                selected=resolved,
                background=semantic["background"],
            )
        if preferred is None:
            preferred = _fallback_distinct_color(
                index=index,
                semantic=semantic,
                quantize_ansi256=quantize_ansi256,
            )
        resolved.append(preferred)

    variables["selection-selected-background"] = _as_hex(resolved[0])
    variables["selection-active-background"] = _as_hex(resolved[1])
    variables["selection-selected-active-background"] = _as_hex(resolved[2])
    variables["selection-selected-foreground"] = _as_hex(
        _preferred_text_color(resolved[0])
    )
    variables["selection-active-foreground"] = _as_hex(
        _preferred_text_color(resolved[1])
    )
    variables["selection-selected-active-foreground"] = _as_hex(
        _preferred_text_color(resolved[2])
    )


def _parse_variable_color(
    value: str | None,
    fallback: ColorTriplet,
) -> ColorTriplet:
    if not value:
        return fallback
    try:
        return Color.parse(value).get_truecolor()
    except Exception:
        return fallback


def _selection_states_are_distinct(values: list[ColorTriplet]) -> bool:
    return all(
        _distance_sq(values[left], values[right]) >= 900
        for left in range(len(values))
        for right in range(left + 1, len(values))
    )


def _selection_candidates(
    *,
    semantic: Mapping[str, ColorTriplet],
    dark: bool,
    quantize_ansi256: bool,
) -> list[ColorTriplet]:
    seeds = (
        semantic["accent"],
        semantic["primary"],
        semantic["secondary"],
        semantic["warning"],
        semantic["error"],
        semantic["success"],
        semantic["foreground"],
    )
    background = semantic["background"]
    mixes = (0.20, 0.30, 0.40, 0.52, 0.64)
    raw: list[ColorTriplet] = []
    for seed in seeds:
        for mix in mixes:
            color = _blend(background, seed, mix)
            color = _ensure_contrast(
                color,
                background,
                minimum_ratio=1.35,
                prefer_light=not dark,
            )
            if quantize_ansi256:
                color = _nearest_xterm_256(color)
            raw.append(color)
    unique: dict[str, ColorTriplet] = {}
    for color in raw:
        unique[_as_hex(color)] = color
    return list(unique.values())


def _is_color_distinct_from(
    color: ColorTriplet,
    others: list[ColorTriplet],
    *,
    minimum_distance: int,
) -> bool:
    return all(
        _distance_sq(color, other) >= minimum_distance
        for other in others
    )


def _pick_distinct_candidate(
    *,
    candidates: list[ColorTriplet],
    selected: list[ColorTriplet],
    background: ColorTriplet,
) -> ColorTriplet | None:
    best: ColorTriplet | None = None
    best_score = -1.0
    for candidate in candidates:
        if not _is_color_distinct_from(
            candidate,
            selected,
            minimum_distance=900,
        ):
            continue
        if not selected:
            distance_score = 1500.0
        else:
            distance_score = min(
                _distance_sq(candidate, existing)
                for existing in selected
            )
        contrast_score = _contrast_ratio(candidate, background) * 100.0
        score = distance_score + contrast_score
        if score > best_score:
            best = candidate
            best_score = score
    return best


def _fallback_distinct_color(
    *,
    index: int,
    semantic: Mapping[str, ColorTriplet],
    quantize_ansi256: bool,
) -> ColorTriplet:
    background = semantic["background"]
    seed_order = (
        semantic["accent"],
        semantic["primary"],
        semantic["warning"],
    )
    color = _blend(background, seed_order[index % len(seed_order)], 0.55)
    if quantize_ansi256:
        color = _nearest_xterm_256(color)
    return color


def _parse_itermcolors(path: Path) -> TerminalPalette:
    try:
        payload = plistlib.loads(path.read_bytes())
    except Exception as exc:
        raise ValueError(f"Failed to parse plist: {exc}") from exc
    if not isinstance(payload, dict):
        raise ValueError("Top-level plist value must be a mapping")

    ansi = [*list(_DEFAULT_ANSI)]
    for index in range(16):
        key = f"Ansi {index} Color"
        parsed = _triplet_from_iterm_dict(payload.get(key))
        if parsed is not None:
            ansi[index] = parsed

    foreground = _triplet_from_iterm_dict(payload.get("Foreground Color"))
    background = _triplet_from_iterm_dict(payload.get("Background Color"))
    cursor = _triplet_from_iterm_dict(payload.get("Cursor Color"))

    if foreground is None:
        foreground = ansi[15]
    if background is None:
        background = ansi[0]
    if cursor is None:
        cursor = foreground

    name = _theme_name_from_path(path)
    return TerminalPalette(
        name=name,
        source=path,
        foreground=foreground,
        background=background,
        cursor=cursor,
        ansi=tuple(ansi),
    )


def _parse_line_theme(path: Path) -> TerminalPalette:
    entries = _parse_line_entries(path)
    ansi = [*list(_DEFAULT_ANSI)]
    for key, value in entries.items():
        ansi_index = _ansi_index_from_key(key)
        if ansi_index is None:
            continue
        parsed = _parse_color_value(value)
        if parsed is None:
            continue
        ansi[ansi_index] = parsed

    foreground = _pick_first_color(entries, _FOREGROUND_KEYS) or ansi[15]
    background = _pick_first_color(entries, _BACKGROUND_KEYS) or ansi[0]
    cursor = _pick_first_color(entries, _CURSOR_KEYS) or foreground

    name = _theme_name_from_path(path)
    custom_name = entries.get("name")
    if custom_name:
        name = custom_name.strip()

    return TerminalPalette(
        name=name,
        source=path,
        foreground=foreground,
        background=background,
        cursor=cursor,
        ansi=tuple(ansi),
    )


def _parse_line_entries(path: Path) -> dict[str, str]:
    entries: dict[str, str] = {}
    section = ""
    try:
        raw = path.read_text(encoding="utf-8")
    except OSError as exc:
        raise ValueError(f"Failed to read file: {exc}") from exc

    for original_line in raw.splitlines():
        line = original_line.strip()
        if not line or line.startswith("#") or line.startswith(";"):
            continue
        if line.startswith("[") and line.endswith("]"):
            section = _normalize_key(line[1:-1])
            continue

        splitter = "=" if "=" in line else ":" if ":" in line else None
        if splitter is None:
            continue
        key, value = line.split(splitter, 1)
        norm_key = _normalize_key(key)
        if section:
            norm_key = f"{section}{norm_key}"
        cleaned = value.strip().strip('"').strip("'")
        entries[norm_key] = cleaned

    collapsed: dict[str, str] = {}
    for key, value in entries.items():
        collapsed[key] = value
        suffix = re.sub(r"^[a-z]+", "", key)
        if suffix and suffix not in collapsed:
            collapsed[suffix] = value
    return collapsed


def _pick_first_color(
    entries: Mapping[str, str],
    keys: tuple[str, ...],
) -> ColorTriplet | None:
    for key in keys:
        value = entries.get(key)
        if value is None:
            continue
        parsed = _parse_color_value(value)
        if parsed is not None:
            return parsed
    return None


def _parse_color_value(value: str) -> ColorTriplet | None:
    try:
        return Color.parse(value).get_truecolor()
    except Exception:
        pass

    numeric = re.findall(r"[+-]?\d*\.?\d+", value)
    if len(numeric) != 3:
        return None

    values = [float(part) for part in numeric]
    if max(values) <= 1.0:
        return ColorTriplet(
            _clamp_byte(round(values[0] * 255)),
            _clamp_byte(round(values[1] * 255)),
            _clamp_byte(round(values[2] * 255)),
        )

    return ColorTriplet(
        _clamp_byte(round(values[0])),
        _clamp_byte(round(values[1])),
        _clamp_byte(round(values[2])),
    )


def _triplet_from_iterm_dict(value: object) -> ColorTriplet | None:
    if not isinstance(value, dict):
        return None
    red = _component_to_byte(value.get("Red Component"))
    green = _component_to_byte(value.get("Green Component"))
    blue = _component_to_byte(value.get("Blue Component"))
    if red is None or green is None or blue is None:
        return None
    return ColorTriplet(red, green, blue)


def _component_to_byte(value: object) -> int | None:
    if not isinstance(value, (float, int)):
        return None
    channel = float(value)
    if channel < 0:
        return None
    if channel <= 1.0:
        return _clamp_byte(round(channel * 255))
    if channel <= 255.0:
        return _clamp_byte(round(channel))
    if channel <= 65535.0:
        return _clamp_byte(round(channel / 257))
    return None


def _pick_slot_color(
    palette: TerminalPalette,
    slots: tuple[int, ...],
    background: ColorTriplet,
    *,
    minimum_ratio: float,
    prefer_light: bool,
) -> ColorTriplet:
    candidates = [
        palette.ansi[slot]
        for slot in slots
        if slot < len(palette.ansi)
    ]
    if not candidates:
        candidates = [palette.foreground]

    best = max(
        candidates,
        key=lambda color: _color_score(color, background, minimum_ratio),
    )
    return _ensure_contrast(
        best,
        background,
        minimum_ratio=minimum_ratio,
        prefer_light=prefer_light,
    )


def _color_score(
    color: ColorTriplet,
    background: ColorTriplet,
    minimum_ratio: float,
) -> float:
    contrast = _contrast_ratio(color, background)
    chroma = _chroma(color) / 255.0
    contrast_bonus = 1.0 if contrast >= minimum_ratio else 0.0
    return contrast + (0.4 * chroma) + contrast_bonus


def _chroma(color: ColorTriplet) -> int:
    highest = max(color.red, color.green, color.blue)
    lowest = min(color.red, color.green, color.blue)
    return highest - lowest


def _derive_theme_variables(
    semantic: Mapping[str, ColorTriplet],
    *,
    action_mix: float,
    selection_mix: float,
    active_mix: float,
    button_mix: float,
    dark: bool,
) -> dict[str, str]:
    background = semantic["background"]
    foreground = semantic["foreground"]
    panel = semantic["panel"]
    accent = semantic["accent"]
    primary = semantic["primary"]

    top_action_background = _blend(panel, foreground, action_mix)
    hover_background = _blend(top_action_background, foreground, 0.16)
    mnemonic_foreground = _ensure_contrast(
        accent,
        top_action_background,
        minimum_ratio=4.5,
        prefer_light=dark,
    )
    selection_background = _blend(background, accent, selection_mix)
    selection_active = _blend(background, primary, active_mix)

    selected_foreground = _preferred_text_color(selection_background)
    active_foreground = _preferred_text_color(selection_active)

    return {
        **CYBER_THEME_VARIABLE_DEFAULTS,
        "top-actions-background": _as_hex(panel),
        "top-action-background": _as_hex(top_action_background),
        "top-action-hover-background": _as_hex(hover_background),
        "top-action-mnemonic-foreground": _as_hex(mnemonic_foreground),
        "selection-selected-background": _as_hex(selection_background),
        "selection-selected-foreground": _as_hex(selected_foreground),
        "selection-active-background": _as_hex(selection_active),
        "selection-active-foreground": _as_hex(active_foreground),
        "selection-selected-active-background": _as_hex(
            _blend(selection_background, selection_active, 0.5)
        ),
        "selection-selected-active-foreground": _as_hex(
            _preferred_text_color(
                _blend(selection_background, selection_active, 0.5)
            )
        ),
        "action-button-background": _as_hex(
            _blend(panel, accent, button_mix)
        ),
        "action-button-foreground": _as_hex(
            _preferred_text_color(_blend(panel, accent, button_mix))
        ),
        "action-button-hover-background": _as_hex(
            _blend(panel, accent, min(1.0, button_mix + 0.12))
        ),
        "action-button-focus-background": _as_hex(
            _blend(panel, accent, min(1.0, button_mix + 0.20))
        ),
        "context-menu-background": _as_hex(panel),
        "context-menu-border": _as_hex(
            _blend(accent, foreground, 0.35 if dark else 0.20)
        ),
    }


def _apply_overrides(
    semantic: dict[str, ColorTriplet],
    variables: dict[str, str],
    overrides: Mapping[str, str],
    palette: TerminalPalette,
) -> None:
    for key, raw_value in overrides.items():
        normalized_key = key.strip().lower()
        color = _resolve_override_color(raw_value, semantic, palette)
        if color is None:
            continue
        if normalized_key in _SEMANTIC_COLOR_KEYS:
            semantic[normalized_key] = color
            continue
        if normalized_key in variables:
            variables[normalized_key] = _as_hex(color)


def _resolve_override_color(
    raw_value: str,
    semantic: Mapping[str, ColorTriplet],
    palette: TerminalPalette,
) -> ColorTriplet | None:
    value = raw_value.strip()
    lower = value.lower()

    semantic_match = semantic.get(lower)
    if semantic_match is not None:
        return semantic_match

    if lower in {"bg", "background"}:
        return palette.background
    if lower in {"fg", "foreground"}:
        return palette.foreground
    if lower == "cursor":
        return palette.cursor

    ansi_match = re.fullmatch(r"ansi(\d{1,2})", lower)
    if ansi_match:
        index = int(ansi_match.group(1))
        if 0 <= index < len(palette.ansi):
            return palette.ansi[index]

    try:
        return Color.parse(value).get_truecolor()
    except Exception:
        return None


def _normalize_override_map(
    overrides: Mapping[str, Mapping[str, str]] | None,
) -> dict[str, dict[str, str]]:
    normalized: dict[str, dict[str, str]] = {}
    if overrides is None:
        return normalized

    for theme_name, mapping in overrides.items():
        if not isinstance(theme_name, str) or not isinstance(mapping, Mapping):
            continue
        values: dict[str, str] = {}
        for key, value in mapping.items():
            if not isinstance(key, str) or not isinstance(value, str):
                continue
            values[key.strip()] = value.strip()
        if values:
            normalized[theme_name.strip()] = values
    return normalized


def _unique_theme_name(name: str, existing: set[str]) -> str:
    candidate = name
    suffix = 2
    while candidate in existing:
        candidate = f"{name} ({suffix})"
        suffix += 1
    existing.add(candidate)
    return candidate


def _theme_name_from_path(path: Path) -> str:
    stem = path.stem.replace("_", " ").strip()
    if not stem:
        return path.name
    return stem


def _ansi_index_from_key(key: str) -> int | None:
    normalized = _normalize_key(key)
    match = re.search(r"(?:ansi|color)(\d{1,2})(?:color)?$", normalized)
    if match:
        value = int(match.group(1))
        if 0 <= value < 16:
            return value

    return _NAMED_SLOT_KEYS.get(normalized)


def _normalize_key(value: str) -> str:
    return re.sub(r"[^a-z0-9]", "", value.strip().lower())


def _blend(
    start: ColorTriplet,
    end: ColorTriplet,
    amount: float,
) -> ColorTriplet:
    weight = max(0.0, min(1.0, amount))
    red = round(start.red + (end.red - start.red) * weight)
    green = round(start.green + (end.green - start.green) * weight)
    blue = round(start.blue + (end.blue - start.blue) * weight)
    return ColorTriplet(red, green, blue)


def _preferred_text_color(background: ColorTriplet) -> ColorTriplet:
    white = ColorTriplet(255, 255, 255)
    black = ColorTriplet(0, 0, 0)
    if _contrast_ratio(white, background) >= _contrast_ratio(
        black,
        background,
    ):
        return white
    return black


def _linear_channel(channel: int) -> float:
    scaled = channel / 255
    if scaled <= 0.03928:
        return scaled / 12.92
    return ((scaled + 0.055) / 1.055) ** 2.4


def _luminance(color: ColorTriplet) -> float:
    return (
        (0.2126 * _linear_channel(color.red))
        + (0.7152 * _linear_channel(color.green))
        + (0.0722 * _linear_channel(color.blue))
    )


def _contrast_ratio(left: ColorTriplet, right: ColorTriplet) -> float:
    left_luma = _luminance(left)
    right_luma = _luminance(right)
    light = max(left_luma, right_luma)
    dark = min(left_luma, right_luma)
    return (light + 0.05) / (dark + 0.05)


def _ensure_contrast(
    color: ColorTriplet,
    background: ColorTriplet,
    *,
    minimum_ratio: float,
    prefer_light: bool,
) -> ColorTriplet:
    if _contrast_ratio(color, background) >= minimum_ratio:
        return color
    target = ColorTriplet(255, 255, 255)
    if not prefer_light:
        target = ColorTriplet(0, 0, 0)
    for step in range(1, 21):
        candidate = _blend(color, target, step / 20)
        if _contrast_ratio(candidate, background) >= minimum_ratio:
            return candidate
    return target


def _as_hex(color: ColorTriplet) -> str:
    return "#{:02x}{:02x}{:02x}".format(
        color.red,
        color.green,
        color.blue,
    )


def _clamp_byte(channel: int) -> int:
    return max(0, min(255, channel))


def _nearest_xterm_256(color: ColorTriplet) -> ColorTriplet:
    return min(
        _XTERM_256,
        key=lambda candidate: _distance_sq(color, candidate),
    )


def _distance_sq(left: ColorTriplet, right: ColorTriplet) -> int:
    red = left.red - right.red
    green = left.green - right.green
    blue = left.blue - right.blue
    return (red * red) + (green * green) + (blue * blue)


def _build_xterm_256_palette() -> tuple[ColorTriplet, ...]:
    palette = list(_DEFAULT_ANSI)
    steps = (0, 95, 135, 175, 215, 255)
    for red in steps:
        for green in steps:
            for blue in steps:
                palette.append(ColorTriplet(red, green, blue))
    for gray in range(8, 239, 10):
        palette.append(ColorTriplet(gray, gray, gray))
    return tuple(palette)


_XTERM_256 = _build_xterm_256_palette()


_BACKGROUND_KEYS = (
    "background",
    "bg",
    "backgroundcolor",
    "terminalbackground",
    "colorsbackground",
    "generalbackground",
)

_FOREGROUND_KEYS = (
    "foreground",
    "fg",
    "foregroundcolor",
    "textcolor",
    "generalforeground",
    "generaltext",
)

_CURSOR_KEYS = (
    "cursor",
    "cursorcolor",
    "cursorforeground",
)

_NAMED_SLOT_KEYS = {
    "black": 0,
    "red": 1,
    "green": 2,
    "yellow": 3,
    "blue": 4,
    "magenta": 5,
    "purple": 5,
    "cyan": 6,
    "white": 7,
    "brightblack": 8,
    "brightred": 9,
    "brightgreen": 10,
    "brightyellow": 11,
    "brightblue": 12,
    "brightmagenta": 13,
    "brightpurple": 13,
    "brightcyan": 14,
    "brightwhite": 15,
    "lightblack": 8,
    "lightred": 9,
    "lightgreen": 10,
    "lightyellow": 11,
    "lightblue": 12,
    "lightmagenta": 13,
    "lightpurple": 13,
    "lightcyan": 14,
    "lightwhite": 15,
}


def default_theme_store_dir(config_path: Path | None = None) -> Path:
    """Return the directory used to store converted themes."""

    _ = config_path
    return Path.home() / ".config" / "sm-logtool" / "themes"


def default_theme_source_dir(config_path: Path | None = None) -> Path:
    """Return the default directory containing import source theme files."""

    _ = config_path
    return Path.home() / ".config" / "sm-logtool" / "theme-sources"


def ensure_default_theme_dirs(
    config_path: Path | None = None,
) -> tuple[Path, Path]:
    """Ensure default per-user theme directories exist and return them."""

    store_dir = default_theme_store_dir(config_path).expanduser()
    source_dir = default_theme_source_dir(config_path).expanduser()
    for directory in (store_dir, source_dir):
        directory.mkdir(parents=True, exist_ok=True)
    return store_dir, source_dir


def save_converted_theme(
    *,
    theme: Theme,
    store_dir: Path,
    source_path: Path,
    mapping_profile: str,
    quantize_ansi256: bool,
) -> Path:
    """Persist a converted theme for future reuse by the TUI."""

    normalized = normalize_mapping_profile(mapping_profile)
    directory = store_dir.expanduser()
    directory.mkdir(parents=True, exist_ok=True)
    target = _theme_path_for_name(directory, theme.name)
    payload = {
        "name": theme.name,
        "dark": bool(theme.dark),
        "primary": theme.primary,
        "secondary": theme.secondary,
        "warning": theme.warning,
        "error": theme.error,
        "success": theme.success,
        "accent": theme.accent,
        "foreground": theme.foreground,
        "background": theme.background,
        "surface": theme.surface,
        "panel": theme.panel,
        "variables": dict(theme.variables or {}),
        "meta": {
            "source_path": str(source_path.expanduser()),
            "mapping_profile": normalized,
            "quantize_ansi256": bool(quantize_ansi256),
            "saved_at_utc": datetime.now(timezone.utc).isoformat(),
        },
    }
    with target.open("w", encoding="utf-8") as handle:
        yaml.safe_dump(payload, handle, sort_keys=False)
    return target


def load_saved_themes(
    *,
    store_dir: Path,
    existing_names: set[str] | None = None,
) -> tuple[list[Theme], list[str]]:
    """Load converted themes previously saved with ``save_converted_theme``."""

    _ = existing_names
    files = _discover_saved_theme_files(store_dir.expanduser())
    themes: list[Theme] = []
    warnings: list[str] = []
    for path in files:
        try:
            payload = _load_yaml_mapping(path)
            theme = _theme_from_payload(payload)
        except ValueError as exc:
            warnings.append(f"{path}: {exc}")
            continue
        themes.append(theme)
    return themes, warnings


def _discover_saved_theme_files(store_dir: Path) -> list[Path]:
    if not store_dir.exists() or not store_dir.is_dir():
        return []
    files = [
        path
        for path in store_dir.iterdir()
        if path.is_file() and path.name.endswith(THEME_FILE_SUFFIX)
    ]
    return sorted(files)


def _theme_path_for_name(store_dir: Path, theme_name: str) -> Path:
    slug = _slugify(theme_name)
    if not slug:
        slug = "theme"
    return store_dir / f"{slug}{THEME_FILE_SUFFIX}"


def _slugify(value: str) -> str:
    slug = re.sub(r"[^a-z0-9]+", "-", value.strip().lower())
    return slug.strip("-")


def _load_yaml_mapping(path: Path) -> dict[str, object]:
    try:
        with path.open("r", encoding="utf-8") as handle:
            loaded = yaml.safe_load(handle) or {}
    except OSError as exc:
        raise ValueError(f"Failed to read YAML: {exc}") from exc
    except yaml.YAMLError as exc:
        raise ValueError(f"Failed to parse YAML: {exc}") from exc
    if not isinstance(loaded, dict):
        raise ValueError("Theme file must contain a mapping.")
    return loaded


def _theme_from_payload(payload: Mapping[str, object]) -> Theme:
    name = _require_string(payload, "name")
    dark = payload.get("dark")
    if not isinstance(dark, bool):
        raise ValueError("Theme key 'dark' must be a boolean.")
    variables = payload.get("variables", {})
    if not isinstance(variables, dict):
        raise ValueError("Theme key 'variables' must be a mapping.")
    variables_dict: dict[str, str] = {}
    for key, value in variables.items():
        if not isinstance(key, str) or not isinstance(value, str):
            raise ValueError("Theme variables must be string key/value pairs.")
        variables_dict[key] = value

    return Theme(
        name=name,
        primary=_require_string(payload, "primary"),
        secondary=_require_string(payload, "secondary"),
        warning=_require_string(payload, "warning"),
        error=_require_string(payload, "error"),
        success=_require_string(payload, "success"),
        accent=_require_string(payload, "accent"),
        foreground=_require_string(payload, "foreground"),
        background=_require_string(payload, "background"),
        surface=_require_string(payload, "surface"),
        panel=_require_string(payload, "panel"),
        dark=dark,
        variables=variables_dict,
    )


def _require_string(payload: Mapping[str, object], key: str) -> str:
    value = payload.get(key)
    if not isinstance(value, str) or not value:
        raise ValueError(f"Theme key '{key}' must be a non-empty string.")
    return value
